import numpy

def xavier_initialization(shape):
    matrix = numpy.random.randn(*shape) / numpy.sqrt(shape[-2])
    return matrix


def he_initialization(shape):
    matrix = numpy.random.randn(*shape) / numpy.sqrt(shape[-2] / 2)
    return matrix


def bias_initialization(shape):
    matrix = numpy.random.normal(0, 1, shape) / shape[-1]
    return matrix


def normal_initialization(shape):
    matrix = numpy.random.normal(0, 1, shape)
    return matrix


def uniform_initialization(shape):
    matrix = numpy.random.randn(*shape)
    return matrix



#-*- conding:utf-8 -*-
import matplotlib.pyplot as plt
import trik
# 每类随机生成点的个数。
points_sum = 100

# 在(0, 0)点附近生成一堆点然后以4为半径在周围生成一堆点构成2类随机数据。
c1_x, c1_y, c2_x, c2_y = [], [], [], []
for _ in range(points_sum):
    c1_x.append(trik.random.normal(0, 1))
    c1_y.append(trik.random.normal(0, 1))
    r = trik.random.normal(4, 1)
    theta = trik.random.normal(0,6.28)
    c2_x.append(r * trik.cos(theta))
    c2_y.append(r * trik.sin(theta))
c_x = c1_x + c2_x
c_y = c1_y + c2_y

# 定义符号。
A = trik.variable([c_x, c_y], name='A')
W1 = trik.variable(trik.random.random((32, 2)), name='W1')  # 输入层到隐含层的权重矩阵。
W2 = trik.variable(trik.random.random((2, 32)), name='W2')  # 隐含层到输出层的权重矩阵。
B1 = trik.variable(trik.random.random((32, 1)), name='B1')  # 隐含层的偏置。
B2 = trik.variable(trik.random.random((2, 1)), name='B2')  # 输出层的偏置。
K = trik.constant([[-1] * points_sum + [1] * points_sum, [1] * points_sum + [-1] * points_sum])

# 构建2x4x2网络，使用ReLu激活函数。
model = W2(trik.MM(),W1(trik.MM(),A)(trik.Plus(),B1)(trik.Max(),0))(trik.Plus(),B2)(trik.Max(),0)
# 使用SVM Pluss。
loss =K(trik.Mul(),model)(trik.ReduceSum(axis=0))(trik.Plus(),1)(trik.Max(),0)(trik.ReduceMean())
print(loss)
# 创建lossPlus擎，申明变量为W1，W2，B1和B2。
loss.variables=[W1, W2, B1, B2]

# 创建梯度下降optimizer。
optimizer = trik.SGD()
# 迭代至多10000次最小化loss。

    
for epoch in range(1000):
    optimizer.minimize(loss)
    if epoch % 10 == 0:  # 每100次epoch检查一次loss。
        loss_value = loss.val()
        print('loss = {:.8f}'.format(loss_value))
        if loss_value < 0.001:  # loss阈值。
            break

# 创建预测函数。
predict = trik.where(model(trik.Mul(),[[-1], [1]])(trik.ReduceSum(axis=0))(trik.Lt(),0), -1, 1)

# 创建预测函数计算引擎。


# 设置网格密度为0.1。
h = 0.1

# 生成预测采样点网格。
x, y = trik.meshgrid(trik.arange(min(c_x) - 1, max(c_x) + 1, h), trik.arange(min(c_y) - 1, max(c_y) + 1, h))

# 绑定变量值。
predict.bind = {A: [x.ravel(), y.ravel()]}

# 生成采样点预测值。
z = predict.val().reshape(x.shape)

# 绘制图像。
plt.title('Paradox implement 2x4x2 Neural Network')
plt.plot(c1_x, c1_y, 'ro', label='Category 1')
plt.plot(c2_x, c2_y, 'bo', label='Category 2')
plt.contourf(x, y, z, 2, cmap='RdBu', alpha=.6)
plt.legend()
plt.show()



import numpy

def helical_data(number: int,
                 category: int=2,
                 noise: float=0,
                 max_radius: int=5,
                 max_theta: float=3*numpy.pi,
                 init_radius: float=0,
                 init_theta: float=0,
                 center: tuple=(0, 0)):
    radius_step = max_radius / number
    theta_step = max_theta / number
    phase = 2 * numpy.pi / category
    data = [[[], []] for _ in range(category)]
    start_phase = [i * phase for i in range(category)]
    radius = init_radius
    theta = init_theta
    x0, x1 = center
    for _ in range(number):
        for i in range(category):
            data[i][0].append(x0 + radius * numpy.cos(theta + start_phase[i]) + numpy.random.normal(0, noise))
            data[i][1].append(x1 + radius * numpy.sin(theta + start_phase[i]) + numpy.random.normal(0, noise))
        radius += radius_step
        theta += theta_step
    return data


def grid_data(number: int,
              noise: float=0,
              raw: int=2,
              column: int=2,
              width: float=5,
              height: float=5,
              ):
    data = [[[], []], [[], []]]
    raw_step = height / raw
    column_step = width / column
    for _ in range(number):
        for i in range(raw):
            for j in range(column):
                x = numpy.random.random() * column_step
                y = numpy.random.random() * raw_step
                data[(i + j) % 2][0].append(x + column_step * j + numpy.random.normal(0, noise))
                data[(i + j) % 2][1].append(y + raw_step * i + numpy.random.normal(0, noise))
    return data


def circle_data(number: int,
                category: int=2,
                center: tuple=(0, 0),
                delta_radius: float=1.,
                noise: float=0.
                ):
    data = [[[], []] for _ in range(category)]
    radius = 0.
    theta_step = 2. * numpy.pi / number
    x0, x1 = center
    for i in range(category):
        radius += delta_radius
        theta = 0.
        for _ in range(number):
            data[i][0].append(x0 + radius * numpy.cos(theta) + numpy.random.normal(0, noise))
            data[i][1].append(x1 + radius * numpy.sin(theta) + numpy.random.normal(0, noise))
            theta += theta_step
    return data


def gaussian_data(number: int,
                  category: int=2,
                  mean: list=None,
                  cov: list=None):
    data = []
    mean = mean or [[i + 1, i + 1] for i in range(category)]
    cov = cov or [numpy.eye(2) for _ in range(category)]
    for i in range(category):
        data.append(numpy.random.multivariate_normal(mean[i], cov[i], number).T.tolist())
    return data


def cross_data(number: int,
               category: int=2,
               center: tuple=(0, 0),
               radius: float=2,
               init_angle: float=0.,
               noise: float=0.):
    delta_angle = numpy.pi / category
    data = [[[], []] for _ in range(category)]
    x0, x1 = center
    for i in range(category):
        x_step = 2. * radius * numpy.cos(init_angle) / number
        y_step = 2. * radius * numpy.sin(init_angle) / number
        x_start = x0 + radius * numpy.cos(init_angle)
        y_start = x1 + radius * numpy.sin(init_angle)
        for _ in range(number):
            data[i][0].append(x_start + numpy.random.normal(0, noise))
            data[i][1].append(y_start + numpy.random.normal(0, noise))
            x_start -= x_step
            y_start -= y_step
        init_angle += delta_angle
    return data



#!python3
"""
Python 3 wrapper for identifying objects in images

Requires DLL compilation

Both the GPU and no-GPU version should be compiled; the no-GPU version should be renamed "yolo_cpp_dll_nogpu.dll".

On a GPU system, you can force CPU evaluation by any of:

- Set global variable DARKNET_FORCE_CPU to True
- Set environment variable CUDA_VISIBLE_DEVICES to -1
- Set environment variable "FORCE_CPU" to "true"


To use, either run performDetect() after import, or modify the end of this file.

See the docstring of performDetect() for parameters.

Directly viewing or returning bounding-boxed images requires scikit-image to be installed (`pip install scikit-image`)


Original *nix 2.7: https://github.com/pjreddie/darknet/blob/0f110834f4e18b30d5f101bf8f1724c34b7b83db/python/darknet.py
Windows Python 2.7 version: https://github.com/AlexeyAB/darknet/blob/fc496d52bf22a0bb257300d3c79be9cd80e722cb/build/darknet/x64/darknet.py

@author: Philip Kahn
@date: 20180503
"""
#pylint: disable=R, W0401, W0614, W0703
from ctypes import *
import math
import random
import os

def sample(probs):
    s = sum(probs)
    probs = [a/s for a in probs]
    r = random.uniform(0, 1)
    for i in range(len(probs)):
        r = r - probs[i]
        if r <= 0:
            return i
    return len(probs)-1

def c_array(ctype, values):
    arr = (ctype*len(values))()
    arr[:] = values
    return arr

class BOX(Structure):
    _fields_ = [("x", c_float),
                ("y", c_float),
                ("w", c_float),
                ("h", c_float)]

class DETECTION(Structure):
    _fields_ = [("bbox", BOX),
                ("classes", c_int),
                ("prob", POINTER(c_float)),
                ("mask", POINTER(c_float)),
                ("objectness", c_float),
                ("sort_class", c_int),
                ("uc", POINTER(c_float)),
                ("points", c_int)]

class DETNUMPAIR(Structure):
    _fields_ = [("num", c_int),
                ("dets", POINTER(DETECTION))]

class IMAGE(Structure):
    _fields_ = [("w", c_int),
                ("h", c_int),
                ("c", c_int),
                ("data", POINTER(c_float))]

class METADATA(Structure):
    _fields_ = [("classes", c_int),
                ("names", POINTER(c_char_p))]



#lib = CDLL("/home/pjreddie/documents/darknet/libdarknet.so", RTLD_GLOBAL)
#lib = CDLL("libdarknet.so", RTLD_GLOBAL)
hasGPU = True
cwd = os.path.dirname(__file__)
if os.name == "nt":    
    os.environ['PATH'] = cwd + ';' + os.environ['PATH']
    winGPUdll = os.path.join(cwd, "yolo_cpp_dll.dll")
    winNoGPUdll = os.path.join(cwd, "yolo_cpp_dll_nogpu.dll")
    envKeys = list()
    for k, v in os.environ.items():
        envKeys.append(k)
    try:
        try:
            tmp = os.environ["FORCE_CPU"].lower()
            if tmp in ["1", "true", "yes", "on"]:
                raise ValueError("ForceCPU")
            else:
                print("Flag value '"+tmp+"' not forcing CPU mode")
        except KeyError:
            # We never set the flag
            if 'CUDA_VISIBLE_DEVICES' in envKeys:
                if int(os.environ['CUDA_VISIBLE_DEVICES']) < 0:
                    raise ValueError("ForceCPU")
            try:
                global DARKNET_FORCE_CPU
                if DARKNET_FORCE_CPU:
                    raise ValueError("ForceCPU")
            except NameError:
                pass
            # print(os.environ.keys())
            # print("FORCE_CPU flag undefined, proceeding with GPU")
        if not os.path.exists(winGPUdll):
            raise ValueError("NoDLL")
        lib = CDLL(winGPUdll, RTLD_GLOBAL)
    except (KeyError, ValueError):
        hasGPU = False
        if os.path.exists(winNoGPUdll):
            lib = CDLL(winNoGPUdll, RTLD_GLOBAL)
            print("Notice: CPU-only mode")
        else:
            # Try the other way, in case no_gpu was
            # compile but not renamed
            lib = CDLL(winGPUdll, RTLD_GLOBAL)
            print("Environment variables indicated a CPU run, but we didn't find `"+winNoGPUdll+"`. Trying a GPU run anyway.")
else:
    try:
        lib = CDLL(os.path.join(cwd, "libdarknet.so"), RTLD_GLOBAL)
    except:
        lib=None
if(lib):
    lib.network_width.argtypes = [c_void_p]
    lib.network_width.restype = c_int
    lib.network_height.argtypes = [c_void_p]
    lib.network_height.restype = c_int

    copy_image_from_bytes = lib.copy_image_from_bytes
    copy_image_from_bytes.argtypes = [IMAGE,c_char_p]

    def network_width(net):
        return lib.network_width(net)

    def network_height(net):
        return lib.network_height(net)

    predict = lib.network_predict_ptr
    predict.argtypes = [c_void_p, POINTER(c_float)]
    predict.restype = POINTER(c_float)

    if hasGPU:
        set_gpu = lib.cuda_set_device
        set_gpu.argtypes = [c_int]
        
    init_cpu = lib.init_cpu

    make_image = lib.make_image
    make_image.argtypes = [c_int, c_int, c_int]
    make_image.restype = IMAGE

    get_network_boxes = lib.get_network_boxes
    get_network_boxes.argtypes = [c_void_p, c_int, c_int, c_float, c_float, POINTER(c_int), c_int, POINTER(c_int), c_int]
    get_network_boxes.restype = POINTER(DETECTION)

    make_network_boxes = lib.make_network_boxes
    make_network_boxes.argtypes = [c_void_p]
    make_network_boxes.restype = POINTER(DETECTION)

    free_detections = lib.free_detections
    free_detections.argtypes = [POINTER(DETECTION), c_int]

    free_batch_detections = lib.free_batch_detections
    free_batch_detections.argtypes = [POINTER(DETNUMPAIR), c_int]

    free_ptrs = lib.free_ptrs
    free_ptrs.argtypes = [POINTER(c_void_p), c_int]

    network_predict = lib.network_predict_ptr
    network_predict.argtypes = [c_void_p, POINTER(c_float)]

    reset_rnn = lib.reset_rnn
    reset_rnn.argtypes = [c_void_p]

    load_net = lib.load_network
    load_net.argtypes = [c_char_p, c_char_p, c_int]
    load_net.restype = c_void_p

    load_net_custom = lib.load_network_custom
    load_net_custom.argtypes = [c_char_p, c_char_p, c_int, c_int]
    load_net_custom.restype = c_void_p

    do_nms_obj = lib.do_nms_obj
    do_nms_obj.argtypes = [POINTER(DETECTION), c_int, c_int, c_float]

    do_nms_sort = lib.do_nms_sort
    do_nms_sort.argtypes = [POINTER(DETECTION), c_int, c_int, c_float]

    free_image = lib.free_image
    free_image.argtypes = [IMAGE]

    letterbox_image = lib.letterbox_image
    letterbox_image.argtypes = [IMAGE, c_int, c_int]
    letterbox_image.restype = IMAGE

    load_meta = lib.get_metadata
    lib.get_metadata.argtypes = [c_char_p]
    lib.get_metadata.restype = METADATA

    load_image = lib.load_image_color
    load_image.argtypes = [c_char_p, c_int, c_int]
    load_image.restype = IMAGE

    rgbgr_image = lib.rgbgr_image
    rgbgr_image.argtypes = [IMAGE]

    predict_image = lib.network_predict_image
    predict_image.argtypes = [c_void_p, IMAGE]
    predict_image.restype = POINTER(c_float)

    predict_image_letterbox = lib.network_predict_image_letterbox
    predict_image_letterbox.argtypes = [c_void_p, IMAGE]
    predict_image_letterbox.restype = POINTER(c_float)

    network_predict_batch = lib.network_predict_batch
    network_predict_batch.argtypes = [c_void_p, IMAGE, c_int, c_int, c_int,
                                       c_float, c_float, POINTER(c_int), c_int, c_int]
    network_predict_batch.restype = POINTER(DETNUMPAIR)

    def array_to_image(arr):
        import numpy as np
        # need to return old values to avoid python freeing memory
        arr = arr.transpose(2,0,1)
        c = arr.shape[0]
        h = arr.shape[1]
        w = arr.shape[2]
        arr = np.ascontiguousarray(arr.flat, dtype=np.float32) / 255.0
        data = arr.ctypes.data_as(POINTER(c_float))
        im = IMAGE(w,h,c,data)
        return im, arr

    def classify(net, meta, im):
        out = predict_image(net, im)
        res = []
        for i in range(meta.classes):
            if altNames is None:
                nameTag = meta.names[i]
            else:
                nameTag = altNames[i]
            res.append((nameTag, out[i]))
        res = sorted(res, key=lambda x: -x[1])
        return res

    def detect(net, meta, image, thresh=.5, hier_thresh=.5, nms=.45, debug= False):
        """
        Performs the meat of the detection
        """
        #pylint: disable= C0321
        im = load_image(image, 0, 0)
        if debug: print("Loaded image")
        ret = detect_image(net, meta, im, thresh, hier_thresh, nms, debug)
        free_image(im)
        if debug: print("freed image")
        return ret

    def detect_image(net, meta, im, thresh=.5, hier_thresh=.5, nms=.45, debug= False):
        #import cv2
        #custom_image_bgr = cv2.imread(image) # use: detect(,,imagePath,)
        #custom_image = cv2.cvtColor(custom_image_bgr, cv2.COLOR_BGR2RGB)
        #custom_image = cv2.resize(custom_image,(lib.network_width(net), lib.network_height(net)), interpolation = cv2.INTER_LINEAR)
        #import scipy.misc
        #custom_image = scipy.misc.imread(image)
        #im, arr = array_to_image(custom_image)		# you should comment line below: free_image(im)
        num = c_int(0)
        if debug: print("Assigned num")
        pnum = pointer(num)
        if debug: print("Assigned pnum")
        predict_image(net, im)
        letter_box = 0
        #predict_image_letterbox(net, im)
        #letter_box = 1
        if debug: print("did prediction")
        #dets = get_network_boxes(net, custom_image_bgr.shape[1], custom_image_bgr.shape[0], thresh, hier_thresh, None, 0, pnum, letter_box) # OpenCV
        dets = get_network_boxes(net, im.w, im.h, thresh, hier_thresh, None, 0, pnum, letter_box)
        if debug: print("Got dets")
        num = pnum[0]
        if debug: print("got zeroth index of pnum")
        if nms:
            do_nms_sort(dets, num, meta.classes, nms)
        if debug: print("did sort")
        res = []
        if debug: print("about to range")
        for j in range(num):
            if debug: print("Ranging on "+str(j)+" of "+str(num))
            if debug: print("Classes: "+str(meta), meta.classes, meta.names)
            for i in range(meta.classes):
                if debug: print("Class-ranging on "+str(i)+" of "+str(meta.classes)+"= "+str(dets[j].prob[i]))
                if dets[j].prob[i] > 0:
                    b = dets[j].bbox
                    if altNames is None:
                        nameTag = meta.names[i]
                    else:
                        nameTag = altNames[i]
                    if debug:
                        print("Got bbox", b)
                        print(nameTag)
                        print(dets[j].prob[i])
                        print((b.x, b.y, b.w, b.h))
                    res.append((nameTag, dets[j].prob[i], (b.x, b.y, b.w, b.h)))
        if debug: print("did range")
        res = sorted(res, key=lambda x: -x[1])
        if debug: print("did sort")
        free_detections(dets, num)
        if debug: print("freed detections")
        return res


    netMain = None
    metaMain = None
    altNames = None

    def performDetect(imagePath=os.path.join(cwd,"dog.jpg"), thresh= 0.25, configPath = os.path.join(cwd,"yolov4.cfg"), weightPath = "yolov4.weights", metaPath= "coco.data", showImage= True, makeImageOnly = False, initOnly= False):
        """
        Convenience function to handle the detection and returns of objects.

        Displaying bounding boxes requires libraries scikit-image and numpy

        Parameters
        ----------------
        imagePath: str
            Path to the image to evaluate. Raises ValueError if not found

        thresh: float (default= 0.25)
            The detection threshold

        configPath: str
            Path to the configuration file. Raises ValueError if not found

        weightPath: str
            Path to the weights file. Raises ValueError if not found

        metaPath: str
            Path to the data file. Raises ValueError if not found

        showImage: bool (default= True)
            Compute (and show) bounding boxes. Changes return.

        makeImageOnly: bool (default= False)
            If showImage is True, this won't actually *show* the image, but will create the array and return it.

        initOnly: bool (default= False)
            Only initialize globals. Don't actually run a prediction.

        Returns
        ----------------------


        When showImage is False, list of tuples like
            ('obj_label', confidence, (bounding_box_x_px, bounding_box_y_px, bounding_box_width_px, bounding_box_height_px))
            The X and Y coordinates are from the center of the bounding box. Subtract half the width or height to get the lower corner.

        Otherwise, a dict with
            {
                "detections": as above
                "image": a numpy array representing an image, compatible with scikit-image
                "caption": an image caption
            }
        """
        # Import the global variables. This lets us instance Darknet once, then just call performDetect() again without instancing again
        global metaMain, netMain, altNames #pylint: disable=W0603
        assert 0 < thresh < 1, "Threshold should be a float between zero and one (non-inclusive)"
        if not os.path.exists(configPath):
            raise ValueError("Invalid config path `"+os.path.abspath(configPath)+"`")
        if not os.path.exists(weightPath):
            raise ValueError("Invalid weight path `"+os.path.abspath(weightPath)+"`")
        if not os.path.exists(metaPath):
            raise ValueError("Invalid data file path `"+os.path.abspath(metaPath)+"`")
        if netMain is None:
            netMain = load_net_custom(configPath.encode("ascii"), weightPath.encode("ascii"), 0, 1)  # batch size = 1
        if metaMain is None:
            metaMain = load_meta(metaPath.encode("ascii"))
        if altNames is None:
            # In Python 3, the metafile default access craps out on Windows (but not Linux)
            # Read the names file and create a list to feed to detect
            try:
                with open(metaPath) as metaFH:
                    metaContents = metaFH.read()
                    import re
                    match = re.search("names *= *(.*)$", metaContents, re.IGNORECASE | re.MULTILINE)
                    if match:
                        result = match.group(1)
                    else:
                        result = None
                    try:
                        if os.path.exists(result):
                            with open(result) as namesFH:
                                namesList = namesFH.read().strip().split("\n")
                                altNames = [x.strip() for x in namesList]
                    except TypeError:
                        pass
            except Exception:
                pass
        if initOnly:
            print("Initialized detector")
            return None
        if not os.path.exists(imagePath):
            raise ValueError("Invalid image path `"+os.path.abspath(imagePath)+"`")
        # Do the detection
        #detections = detect(netMain, metaMain, imagePath, thresh)	# if is used cv2.imread(image)
        detections = detect(netMain, metaMain, imagePath.encode("ascii"), thresh)
        if showImage:
            try:
                from skimage import io, draw
                import numpy as np
                image = io.imread(imagePath)
                print("*** "+str(len(detections))+" Results, color coded by confidence ***")
                imcaption = []
                for detection in detections:
                    label = detection[0]
                    confidence = detection[1]
                    pstring = label+": "+str(np.rint(100 * confidence))+"%"
                    imcaption.append(pstring)
                    print(pstring)
                    bounds = detection[2]
                    shape = image.shape
                    # x = shape[1]
                    # xExtent = int(x * bounds[2] / 100)
                    # y = shape[0]
                    # yExtent = int(y * bounds[3] / 100)
                    yExtent = int(bounds[3])
                    xEntent = int(bounds[2])
                    # Coordinates are around the center
                    xCoord = int(bounds[0] - bounds[2]/2)
                    yCoord = int(bounds[1] - bounds[3]/2)
                    boundingBox = [
                        [xCoord, yCoord],
                        [xCoord, yCoord + yExtent],
                        [xCoord + xEntent, yCoord + yExtent],
                        [xCoord + xEntent, yCoord]
                    ]
                    # Wiggle it around to make a 3px border
                    rr, cc = draw.polygon_perimeter([x[1] for x in boundingBox], [x[0] for x in boundingBox], shape= shape)
                    rr2, cc2 = draw.polygon_perimeter([x[1] + 1 for x in boundingBox], [x[0] for x in boundingBox], shape= shape)
                    rr3, cc3 = draw.polygon_perimeter([x[1] - 1 for x in boundingBox], [x[0] for x in boundingBox], shape= shape)
                    rr4, cc4 = draw.polygon_perimeter([x[1] for x in boundingBox], [x[0] + 1 for x in boundingBox], shape= shape)
                    rr5, cc5 = draw.polygon_perimeter([x[1] for x in boundingBox], [x[0] - 1 for x in boundingBox], shape= shape)
                    boxColor = (int(255 * (1 - (confidence ** 2))), int(255 * (confidence ** 2)), 0)
                    draw.set_color(image, (rr, cc), boxColor, alpha= 0.8)
                    draw.set_color(image, (rr2, cc2), boxColor, alpha= 0.8)
                    draw.set_color(image, (rr3, cc3), boxColor, alpha= 0.8)
                    draw.set_color(image, (rr4, cc4), boxColor, alpha= 0.8)
                    draw.set_color(image, (rr5, cc5), boxColor, alpha= 0.8)
                if not makeImageOnly:
                    io.imshow(image)
                    io.show()
                detections = {
                    "detections": detections,
                    "image": image,
                    "caption": "\n<br/>".join(imcaption)
                }
            except Exception as e:
                print("Unable to show image: "+str(e))
        return detections

    def performBatchDetect(thresh= 0.25, configPath = "./cfg/yolov3.cfg", weightPath = "yolov3.weights", metaPath= "./cfg/coco.data", hier_thresh=.5, nms=.45, batch_size=3):
        import cv2
        import numpy as np
        # NB! Image sizes should be the same
        # You can change the images, yet, be sure that they have the same width and height
        img_samples = ['data/person.jpg', 'data/person.jpg', 'data/person.jpg']
        image_list = [cv2.imread(k) for k in img_samples]

        net = load_net_custom(configPath.encode('utf-8'), weightPath.encode('utf-8'), 0, batch_size)
        meta = load_meta(metaPath.encode('utf-8'))
        pred_height, pred_width, c = image_list[0].shape
        net_width, net_height = (network_width(net), network_height(net))
        img_list = []
        for custom_image_bgr in image_list:
            custom_image = cv2.cvtColor(custom_image_bgr, cv2.COLOR_BGR2RGB)
            custom_image = cv2.resize(
                custom_image, (net_width, net_height), interpolation=cv2.INTER_NEAREST)
            custom_image = custom_image.transpose(2, 0, 1)
            img_list.append(custom_image)

        arr = np.concatenate(img_list, axis=0)
        arr = np.ascontiguousarray(arr.flat, dtype=np.float32) / 255.0
        data = arr.ctypes.data_as(POINTER(c_float))
        im = IMAGE(net_width, net_height, c, data)

        batch_dets = network_predict_batch(net, im, batch_size, pred_width,
                                                    pred_height, thresh, hier_thresh, None, 0, 0)
        batch_boxes = []
        batch_scores = []
        batch_classes = []
        for b in range(batch_size):
            num = batch_dets[b].num
            dets = batch_dets[b].dets
            if nms:
                do_nms_obj(dets, num, meta.classes, nms)
            boxes = []
            scores = []
            classes = []
            for i in range(num):
                det = dets[i]
                score = -1
                label = None
                for c in range(det.classes):
                    p = det.prob[c]
                    if p > score:
                        score = p
                        label = c
                if score > thresh:
                    box = det.bbox
                    left, top, right, bottom = map(int,(box.x - box.w / 2, box.y - box.h / 2,
                                                box.x + box.w / 2, box.y + box.h / 2))
                    boxes.append((top, left, bottom, right))
                    scores.append(score)
                    classes.append(label)
                    boxColor = (int(255 * (1 - (score ** 2))), int(255 * (score ** 2)), 0)
                    cv2.rectangle(image_list[b], (left, top),
                              (right, bottom), boxColor, 2)
            cv2.imwrite(os.path.basename(img_samples[b]),image_list[b])

            batch_boxes.append(boxes)
            batch_scores.append(scores)
            batch_classes.append(classes)
        free_batch_detections(batch_dets, batch_size)
        return batch_boxes, batch_scores, batch_classes    

    if __name__ == "__main__":
        print(performDetect())
        #Uncomment the following line to see batch inference working 
        #print(performBatchDetect())



from functools import reduce
import numpy
#shape-----cs---------
def element_wise_shape(*shape_list):
	broadcast_map = {_shape: [] for _shape in shape_list}
	new_shape = []
	for shape in shape_list:
		if len(shape) > len(new_shape):
			new_shape = list(shape)
	for i in range(-len(new_shape), 0):
		index = len(new_shape) + i
		dimensions = {}
		for shape in shape_list:
			if -i > len(shape):
				broadcast_map[shape].append(-1)
			else:
				broadcast_map[shape].append(0)
				dimensions[shape] = shape[i]
		new_shape[index] = max([_d for _, _d in dimensions.items()])
		for shape, dimension in dimensions.items():
			if dimension != new_shape[index]:
				if dimension == 1:
					broadcast_map[shape][-1] = 1
				else:
					raise ValueError('Can not broadcast these shapes: {}'.format(shape_list))
	return (tuple(new_shape),) + tuple(tuple(broadcast_map[_shape]) for _shape in shape_list)

def matrix_multiply_shape(shape_a, shape_b):
	try:
		if len(shape_a) == 0 or len(shape_b) == 0:
			raise ValueError()
		if len(shape_a) == 1 and len(shape_b) == 1:
			if shape_a[0] == shape_b[0]:
				return (), (), ()
			else:
				raise ValueError()
		if len(shape_a) == 1:
			if shape_a[0] == shape_b[-2]:
				new_shape = list(shape_b)
				del new_shape[-2]
				return tuple(new_shape), (), ()
			else:
				raise ValueError()
		if len(shape_b) == 1:
			if shape_a[-1] == shape_b[0]:
				return shape_a[:-1], (), ()
			else:
				raise ValueError()
		if shape_a[-1] == shape_b[-2]:
			gap = abs(len(shape_a) - len(shape_b))
			if len(shape_a) > len(shape_b):
				if shape_a[gap:-2] != shape_b[:-2]:
					raise ValueError()
				new_shape = list(shape_a)
				broadcast_a = (0,) * len(shape_a)
				broadcast_b = shape_a[:gap] + (0, 0)
			else:
				if shape_b[gap:-2] != shape_a[:-2]:
					raise ValueError()
				new_shape = list(shape_b)
				broadcast_a = shape_b[:gap] + (0, 0)
				broadcast_b = (0,) * len(shape_b)
			new_shape[-1] = shape_b[-1]
			new_shape[-2] = shape_a[-2]
			return tuple(new_shape), broadcast_a, broadcast_b
		else:
			raise ValueError()
	except ValueError:
		raise ValueError('Can not execute matrix multiply with these two shapes: a={}, b={}'.format(shape_a, shape_b))

def reduce_shape(shape_a, axis, invariant):
	if axis is None:
		return (), ()
	else:
		new_shape = list(shape_a)
		if invariant:
			new_shape[axis] = 1
		else:
			del new_shape[axis]
		return tuple(new_shape), ()


def transpose_shape(shape_a, axes):
	if axes is None:
		return tuple(reversed(shape_a)), ()
	else:
		if len(set(axes)) == len(axes):
			if set(axes) == set(range(len(axes))) and len(axes) == len(shape_a):
				new_shape = [0] * len(shape_a)
				for i, d in zip(axes, shape_a):
					new_shape[i] = d
				return tuple(new_shape), ()
			else:
				raise ValueError('Invalid axes for this Shape: shape={}, axes={}'.format(shape_a, axes))
		else:
			raise ValueError('Repeated axis in axes: {}'.format(axes))

def concatenate_shape(axis, *shape_list):
	new_shape = list(shape_list[0])
	shape_len = len(new_shape)
	for shape in shape_list[1:]:
		if len(shape) == len(shape):
			for i in range(shape_len):
				if i == axis:
					new_shape[i] += shape[i]
				else:
					if new_shape[i] != shape[i]:
						raise ValueError('Concatenate shape not match: {}'.format(shape_list))
		else:
			raise ValueError('All shapes must have same dimensions')
	return (tuple(new_shape),) + () * len(shape_list)

def slice_shape(shape_a, slice_list):
	if not isinstance(slice_list, list):
		slice_list = [slice_list]
	new_shape = list(shape_a)
	delete_dimension = 0
	for i, s in enumerate(slice_list):
		index = i - delete_dimension
		if index < len(shape_a):
			if isinstance(s, slice):
				new_shape[index] = len(([0] * shape_a[index])[s])
			elif isinstance(s, int):
				del new_shape[index]
				delete_dimension += 1
			else:
				raise ValueError('Invalid slice type: {}'.format(type(s)))
		else:
			raise ValueError('Shape not match slice: shape={} slice={}'.format(shape_a, slice_list))
	return tuple(new_shape), ()

def rotate90_shape(shape_a, count, axes):
	new_shape = list(shape_a)
	if count % 2 == 0:
		return tuple(new_shape), ()
	else:
		new_shape[axes[0]], new_shape[axes[1]] = new_shape[axes[1]], new_shape[axes[0]]
		return tuple(new_shape), ()

#optim--------
class Optimizer:
	def optimize(self, engine,calculate_function):
		raise ValueError("this is abstract")
	def minimize(self, engine,steps: int=1):
		for _ in range(steps):
			self.optimize(engine,lambda v, g: v - g)
			
class SGD(Optimizer):
	def __init__(self, rate: float=0.01, factor: float=0.9,decay:float=0.0005,consistent: bool=True):
		self.__rate = rate
		self.__factor = factor
		self.__decay=decay
		self.__consistent = consistent
		self.__old_gradient_map = {}
		self.__gradient_engine = V()

	def __repr__(self):
		return '{}(rate={}, consistent={})'.format(self.__class__.__name__, self.__rate, self.__consistent)

	def optimize(self, engine, calculate_function):
		variables = engine.variables
		for variable in variables:
			value_cache = self.__gradient_engine.value_cache
			self.__gradient_engine = engine.bprop(variable)
			self.__gradient_engine.set_bind(engine.get_bind())
			if self.__consistent:
				self.__gradient_engine.value_cache = value_cache
			momentum = self.__gradient_engine.val() + self.__factor * self.__old_gradient_map.get(variable, 0)-self.__decay*variable.value
			self.__old_gradient_map[variable] = momentum
			variable.value = calculate_function(variable.value, self.__rate * momentum)
		engine.modified()
		self.__gradient_engine.modified()


class Adam(Optimizer):
	def __init__(self, rate: float, decay: float=0.9, square_decay: float=0.999, consistent: bool=True):
		self.__rate = rate
		self.__decay = decay
		self.__square_decay = square_decay
		self.__consistent = consistent
		self.__gradient_engine = V()
		self.__estimation_map = {}
		self.__square_estimation_map = {}
		self.__step = 1

	def __repr__(self):
		return '{}(rate={}, decay={}, square_decay={}, consistent={})'.format(self.__class__.__name__, self.__rate, self.__decay, self.__square_decay, self.__consistent)

	def optimize(self, engine, calculate_function):
		
		variables = engine.variables
		for variable in variables:
			
			value_cache = self.__gradient_engine.value_cache
			self.__gradient_engine = engine.bprop(variable)
			self.__gradient_engine.bind=engine.bind
			if self.__consistent:
				self.__gradient_engine.value_cache = value_cache
			current_gradient = self.__gradient_engine.val()
			self.__estimation_map.setdefault(variable, 0)
			self.__square_estimation_map.setdefault(variable, 0)
			self.__estimation_map[variable] = self.__decay * self.__estimation_map[variable] + (1 - self.__decay) * current_gradient
			self.__square_estimation_map[variable] = self.__square_decay * self.__square_estimation_map[variable] + (1 - self.__square_decay) * current_gradient ** 2
			estimation = self.__estimation_map[variable] / (1 - self.__decay ** self.__step)
			square_estimation = self.__square_estimation_map[variable] / (1 - self.__square_decay ** self.__step)
			self.__step += 1
			regularization_value = estimation / (square_estimation + 1e-8) ** 0.5
			variable.value = calculate_function(variable.value, self.__rate * regularization_value)
		engine.modified()
		self.__gradient_engine.modified()


#tensor------------
class VCategory:
	variable = 0
	constant = 1
	placeholder = 2
	operator = 3
	
class V:
	def __init__(self, value=None, shape: tuple=None, name: str=None, operator=None, inputs=None, category: VCategory=None):
		self.__name = None
		self.__input = []
		self.__operator = None
		self.__output = []
		self.__value = None
		self.__shape = None
		self.__scalar = False
		self.__category = None
		# for engin
		self.__variables = set()
		self.__grad_table = {}
		self.__shape2 = {}
		self.__broadcast = {}
		self.__bind ={}
		self.__value_cache = {}
		if isinstance(value, V):
			self.__set_name(value.name)
			self.__set_value(value.value)
			self.__set_shape(value.shape)
			for _input in value.input:
				self.__add_input(_input.clone())
			self.__set_operator(value.__operator)
			for _output in value.output:
				self.__add_output(_output.clone())
			self.__set_category(value.category)
		else:
			self.symbolic_compute(operator, inputs)
			self.__set_value(value)
			self.__set_category(category)
			self.__set_shape(shape)
			self.__set_name(name)
	def __call__(self,op,*arg):
		arg=[self,]+list(arg)
		res=V(operator=op,inputs=as_symbols(arg))
		simp.simplify(res)
		return res
	def __repr__(self):
		if self.__operator is None:
			return self.__name
		else:
			if self.__operator.operator_sign is None:
				arguments = list(map(str, self.input))
				arguments += ['{}={}'.format(k, '\'' + v + '\'' if isinstance(v, str) else v) for k, v in self.__operator.arguments.items()]
				return '{}({})'.format(self.__operator.__class__.__name__, ', '.join(arguments))
			else:
				if len(self.input) < 2:
					return '{}{}'.format(self.__operator.operator_sign, self.input[0])
				else:
					return '({} {} {})'.format(self.input[0], self.__operator.operator_sign, self.input[1])

	def __get_name(self):
		return self.__name

	def __set_name(self, name: str):
		if name is None:
			if self.__value is None:
				self.__name = self.__class__.__name__
			else:
				if self.is_scalar():
					self.__name = str(self.__value)
				else:
					self.__name = '{}{}'.format(self.__class__.__name__, self.__value.shape)
		else:
			self.__name = name

	name = property(__get_name, __set_name)

	def __get_category(self):
		return self.__category

	def __set_category(self, category: VCategory):
		if category is None:
			if self.__category is None:
				self.__category = VCategory.variable
		else:
			if self.__category == VCategory.operator:
				if category != VCategory.operator:
					raise ValueError('Can not change category of Operator V.')
			if category == VCategory.variable:
				self.__category = category
			elif category == VCategory.constant:
				self.__category = category
				if self.__value is None:
					raise ValueError('constant V must have value.')
			elif category == VCategory.placeholder:
				self.__category = category
				if self.__value is not None:
					self.__set_shape(self.__value.shape)
			elif category == VCategory.operator:
				if self.operator is not None:
					self.__category = category
				else:
					raise ValueError('Can not convert other V to Operator V.')
			else:
				raise ValueError('Invalid category: {}'.format(category))

	category = property(__get_category, __set_category)

	def __get_value(self):
		return self.__value

	def __set_value(self, tensor):
		if tensor is not None:
			if self.__category == VCategory.constant:
				raise ValueError('Can not change value for constant.')
			else:
				if self.__operator is None:
					self.__value = numpy.array(tensor, dtype=float)
					self.__scalar = len(self.value.shape) == 0
					self.__set_category(VCategory.variable)
				else:
					raise ValueError('Can not assign value for Operator V.')

	value = property(__get_value, __set_value)

	def __get_shape(self):
		if self.__category == VCategory.placeholder:
			return self.__shape
		else:
			if self.__operator is None:
				return self.__value.shape
			else:
				return self.shape2(self)
#                 raise ValueError('Operator V has no shape.')

	def __set_shape(self, shape: tuple):
		if shape is not None:
			if self.__category == VCategory.placeholder:
				self.__shape = shape
			else:
				raise ValueError('Only Placeholder can set shape.')

	shape = property(__get_shape, __set_shape)

	def __get_operator(self):
		return self.__operator

	def __set_operator(self, operator):
		if operator is not None:
			if isinstance(operator, Operator):
				self.__operator = operator
			else:
				raise ValueError('Operator must be Operator.')

	operator = property(__get_operator, __set_operator)

	def __get_input(self):
		return self.__input

	input = property(__get_input)

	def __add_input(self, symbol):
		if isinstance(symbol, V):
			self.__input.append(symbol)
		else:
			raise ValueError('Input must be V.')

	def __get_output(self):
		return self.__output

	output = property(__get_output)

	def __add_output(self, symbol):
		if isinstance(symbol, V):
			self.__output.append(symbol)
		else:
			raise ValueError('Output must be V.')

	def symbolic_compute(self, operator, inputs):
		if operator is not None and inputs:
			self.__set_operator(operator)
			inputs_count = operator.inputs_count
			if inputs_count is None:
				inputs_count = len(inputs)
			self.__input = []
			self.__scalar = True
			for symbol in inputs[:inputs_count]:
				if isinstance(symbol, V):
					if not symbol.is_scalar():
						self.__scalar = False
					self.__add_input(symbol)
					symbol.__add_output(self)
				else:
					raise ValueError('Input must be V.')
			self.__set_category(VCategory.operator)

	def remove_input(self, symbol):
		new_input = []
		find_input = None
		for each_input in self.__input:
			if hash(each_input) != hash(symbol):
				new_input.append(each_input)
			else:
				find_input = each_input
		self.__input = new_input
		if find_input is not None:
			find_input.remove_output(self)

	def remove_output(self, symbol):
		new_output = []
		find_output = None
		for each_output in self.__output:
			if hash(each_output) != hash(symbol):
				new_output.append(each_output)
			else:
				find_output = each_output
		self.__output = new_output
		if find_output is not None:
			find_output.remove_input(self)

	def clear_input(self):
		for symbol in set(self.__input):
			symbol.remove_output(self)
		self.__input = []

	def clear_output(self):
		for symbol in set(self.__output):
			symbol.remove_input(self)
		self.__output = []

	def clear_operator(self):
		self.clear_input()
		self.__operator = None
		self.__category = VCategory.variable

	def destroy(self):
		self.clear_input()
		self.clear_output()
		self.__value = None
		self.__operator = None

	def is_scalar(self):
		return self.__scalar
	def is_constant(self):
		return self.__category == VCategory.constant
	def is_variable(self):
		return self.__category == VCategory.variable
	def is_placeholder(self):
		return self.__category == VCategory.placeholder
	def is_operator(self):
		return self.__category == VCategory.operator

	def symbolic_hash(self):
		if self.is_operator():
			inputs_symbolic_hash = [each_input.symbolic_hash() for each_input in self.input]
			return '{}({})'.format(self.operator.__class__.__name__, ','.join(inputs_symbolic_hash))
		else:
			return str(hash(self))

	def __hash__(self):
		return id(self)
	def __getitem__(self, item):
		return slice_select(self, item)
	def __setitem__(self, key, value):
		return slice_assign(self, value, key)
	
	def clear(self):
		self.__grad_table = {}
		self.__shape2 = {}
		self.__broadcast = {}
		self.__value_cache = {}
	def get_variables(self):
		return self.__variables

	def set_variables(self, symbol):
		if symbol is None:
			symbol = set()
			symbol_set ={self} 
			while len(symbol_set):
				any_symbol = symbol_set.pop()
				if any_symbol.is_variable():
					symbol.add(any_symbol)
				elif any_symbol.is_operator():
					symbol_set |= set(any_symbol.input)
		old_variables = set(self.__variables)
		if isinstance(symbol, V):
			symbols = {symbol}
		else:
			symbols = set(symbol)
		for symbol in symbols:
			if isinstance(symbol, V) and not symbol.is_operator():
				self.__variables.add(symbol)
			else:
				raise ValueError('Variable must be V.')
		unused_variables = old_variables - self.__variables
		for variable in unused_variables:
			if variable in self.__grad_table:
				del self.__grad_table[variable]

	variables = property(get_variables, set_variables)

	def get_bind(self,):
		return self.__bind

	def set_bind(self,bind_data: dict):
		old_bind = self.__bind
		self.__bind ={}
		need_clear = False
		for s, d in bind_data.items():
			if s.category == VCategory.constant:
				raise ValueError('Can not bind data for constant.')
			d_array = numpy.array(d)
			if s in old_bind:
				if old_bind[s].shape != d_array.shape:
					need_clear = True
			else:
				need_clear = True
			self.__bind[s] = d_array
		if need_clear:
			self.clear()
	bind=property(get_bind,set_bind)
	def get_value_cache(self):
		return self.__value_cache

	def set_value_cache(self, value_cache: dict):
		self.__value_cache = value_cache

	value_cache = property(get_value_cache, set_value_cache)

	def modified(self):
		self.__value_cache = {}

	def __compute_value(self,symbol):
		if not isinstance(symbol,V):raise ValueError('symbol Must be V')
		if not symbol.is_operator():
			if symbol in self.__bind:
				return numpy.array(self.__bind[symbol])
			else:
				if symbol.value is None or symbol.is_placeholder():
					raise ValueError('V must bind data: {}'.format(symbol))
				else:
					return symbol.value
		else:
			if symbol in self.__value_cache:
				return self.__value_cache[symbol]
			else:
				compute_inputs =[self.__compute_value(_s) for _s in symbol.input]
				symbol_value = symbol.operator.f(*compute_inputs)
				self.__value_cache[symbol] =symbol_value
				return symbol_value

	def __build_grad(self, variable):
		if not isinstance(variable,V):
			raise ValueError('variable must be V')
		if hash(self) == hash(variable):
			self.__grad_table[variable] = constant(1)(Broadcast(self.shape2(self)))
			return
		current_operator = None
		index = -1
		for forward in variable.output:
			if forward.is_constant():continue
			if self.bprop(forward) is not None:
				if current_operator != forward.operator:
					current_operator = forward.operator
					index = -1
				gradients = forward.operator.bprop(self, forward, *forward.input)
				for i, _variable in enumerate(forward.input, start=index + 1):
					if hash(_variable) == hash(variable):
						index = i
						break
				if gradients:current_gradient=gradients[index]()
				else:current_gradient=constant(0)
				if forward.operator.auto_reduce:
					invariant = 0
					for i, d in enumerate(self.broadcast(variable, forward)):
						if d > 0:
							current_gradient = current_gradient(ReduceSum(axis=i + invariant, invariant=True))
						elif d < 0:
							current_gradient = current_gradient(ReduceSum(axis=i + invariant, invariant=False))
							invariant -= 1
				if variable not in self.__grad_table:
					self.__grad_table[variable] = current_gradient
				else:
					self.__grad_table[variable] = self.__grad_table[variable](Plus(),current_gradient)
		if variable in self.__grad_table:
			simp.simplify(self.__grad_table[variable])

	def __compute_shape(self, symbol):
		if not isinstance(symbol,V):
			raise ValueError('symbol must be V')
		if not symbol.is_operator():
			if symbol in self.__bind:
				self.__shape2[symbol] = self.__bind[symbol].shape
			else:
				if symbol.shape is None:
					raise ValueError('Placeholder must bind data or set shape: {}'.format(symbol))
				else:
					self.__shape2[symbol] = symbol.shape
		else:
			shape_broadcasts = symbol.operator.shape(*[self.shape2(s) for s in symbol.input])
			shape = shape_broadcasts[0]
			broadcasts = shape_broadcasts[1:]
			self.__shape2[symbol] = shape
			for input_symbol, input_broadcast in zip(symbol.input, broadcasts):
				if sum([abs(d) for d in input_broadcast]) > 0:
					self.__broadcast.setdefault(input_symbol, {})
					self.__broadcast[input_symbol].setdefault(symbol, {})
					self.__broadcast[input_symbol][symbol] = input_broadcast

	def val(self):
		return self.__compute_value(self)

	def differentiate(self):
		for variable in self.__variables:
			if variable not in self.__grad_table:
				self.__build_grad(variable)

	def bprop(self, variable):
		if not isinstance(variable,V):
			raise ValueError('variable must be V')
		if variable not in self.__grad_table:
			self.__build_grad(variable)
		return self.__grad_table.get(variable, None)

	def shape2(self, variable):
		if not isinstance(variable,V):
			raise ValueError('variable must be V')
		if variable not in self.__shape2:
			self.__compute_shape(variable)
		return self.__shape2.get(variable, None)

	def broadcast(self, from_variable, to_variable):
		if not isinstance(from_variable,V) or not isinstance(to_variable,V):
			raise ValueError('variable must be V')
		if from_variable not in self.__broadcast:
			self.__compute_shape(from_variable)
		if from_variable not in self.__broadcast:
			return ()
		else:
			if to_variable not in self.__broadcast[from_variable]:
				return ()
			else:
				return self.__broadcast[from_variable][to_variable]
	
def constant(value=None, shape: tuple=None, name: str=None, operator=None, inputs=None):
	return V(value, shape, name, operator, inputs, VCategory.constant)
def variable(value=None, shape: tuple=None, name: str=None, operator=None, inputs=None):
	return V(value, shape, name, operator, inputs, VCategory.variable)
def placeholder(value=None, shape: tuple=None, name: str=None, operator=None, inputs=None):
	return V(value, shape, name, operator, inputs, VCategory.placeholder)

def as_symbol(thing):
	if isinstance(thing, V):
		return thing
	else:
		return constant(thing)
def as_symbols(things):
	return list(map(as_symbol, things))
def slice_assign(a, b, slice_tuple):
	return V(operator=SliceAssign(slice_tuple), inputs=as_symbols([a, b]))  
def slice_select(a, slice_tuple):
	return V(operator=SliceSelect(slice_tuple), inputs=as_symbols([a]))

#operator---------
class Operator:
	operator_sign = None
	inputs_count = None
	auto_reduce = True
	arguments = {}

	def __repr__(self):
		return '{}({})'.format(self.__class__.__name__, ', '.join(['{}={}'.format(key, value) for key, value in self.arguments.items()]))

	def f(self, *args, **kwargs):
		raise ValueError("this is abstract")

	def bprop(self, *args, **kwargs):
		raise ValueError("this is abstract")

	def shape(self, *args, **kwargs):
		raise ValueError("this is abstract")

class Neg(Operator):
	def __init__(self):
		self.operator_sign = '-'
		self.inputs_count = 1

	def f(self, value_a):
		return -value_a

	def bprop(self, engine, symbol_forward, symbol_a):
		forward = engine.bprop(symbol_forward)
		return [lambda: forward(Neg())]

	def shape(self, shape_a):
		return shape_a, ()


class Abs(Operator):
	def __init__(self):
		self.inputs_count = 1

	def f(self, value_a):
		return numpy.absolute(value_a)

	def bprop(self, engine, symbol_forward, symbol_a):
		forward = engine.bprop(symbol_forward)
		return [lambda: forward(Mul(),V(operator=Where(),inputs=as_symbols([symbol_a(Gt(),0), 1, -1])))]

	def shape(self, shape_a):
		return shape_a, ()


class Plus(Operator):
	def __init__(self):
		self.operator_sign = '+'
		self.inputs_count = 2

	def f(self, value_a, value_b):
		return value_a + value_b

	def bprop(self, engine, symbol_forward, symbol_a, symbol_b):
		forward = engine.bprop(symbol_forward)
		return [lambda: forward(Mul(),1),
				lambda: forward(Mul(),1)]

	def shape(self, shape_a, shape_b):
		return element_wise_shape(shape_a, shape_b)


class Sub(Operator):
	def __init__(self):
		self.operator_sign = '-'
		self.inputs_count = 2

	def f(self, value_a, value_b):
		return value_a - value_b

	def bprop(self, engine, symbol_forward, symbol_a, symbol_b):
		forward = engine.bprop(symbol_forward)
		return [lambda: forward(Mul(),1),
				lambda: forward(Mul(),-1)]

	def shape(self, shape_a, shape_b):
		return element_wise_shape(shape_a, shape_b)


class Mul(Operator):
	def __init__(self):
		self.operator_sign = '*'
		self.inputs_count = 2

	def f(self, value_a, value_b):
		return value_a * value_b

	def bprop(self, engine, symbol_forward, symbol_a, symbol_b):
		forward = engine.bprop(symbol_forward)
		return [lambda: forward(Mul(),symbol_b),
				lambda: forward(Mul(),symbol_a)]

	def shape(self, shape_a, shape_b):
		return element_wise_shape(shape_a, shape_b)


class Div(Operator):
	def __init__(self):
		self.operator_sign = '/'
		self.inputs_count = 2

	def f(self, value_a, value_b):
		return value_a / value_b

	def bprop(self, engine, symbol_forward, symbol_a, symbol_b):
		forward = engine.bprop(symbol_forward)
		return [lambda: forward(Mul(),1)(Div(),symbol_b),
				lambda: forward(Mul(),-1)(Mul(),symbol_a)(Div(),symbol_b(Pow(),2))]

	def shape(self, shape_a, shape_b):
		return element_wise_shape(shape_a, shape_b)


class MM(Operator):
	def __init__(self):
		self.operator_sign = '@'
		self.inputs_count = 2

	def f(self, value_a, value_b):
		if len(value_a.shape) == 0 or len(value_b.shape) == 0:
			return value_a * value_b
		else:
			return value_a @ value_b

	def bprop(self, engine, symbol_forward, symbol_a, symbol_b):
		forward = engine.bprop(symbol_forward)
		shape_a = engine.shape2(symbol_a)
		shape_b = engine.shape2(symbol_b)
		if len(shape_a) >= 2:
			axes_a = tuple(range(len(shape_a) - 2)) + (-1, -2)
		else:
			axes_a = None
		if len(shape_b) >= 2:
			axes_b = tuple(range(len(shape_b) - 2)) + (-1, -2)
		else:
			axes_b = None
		return [lambda: forward(MM(),symbol_b(T(axes=axes_a))),
				lambda: symbol_a(T(axes=axes_b))(MM(),forward)]

	def shape(self, shape_a, shape_b):
		return matrix_multiply_shape(shape_a, shape_b)


class T(Operator):
	def __init__(self, axes=None):
		self.inputs_count = 1
		self.arguments = {'axes': axes}

	def f(self, value_a):
		return numpy.transpose(value_a, axes=self.arguments['axes'])

	def bprop(self, engine, symbol_forward, symbol_a):
		forward = engine.bprop(symbol_forward)
		return [lambda: forward(T(axes=self.arguments['axes']))]

	def shape(self, shape_a):
		return transpose_shape(shape_a, self.arguments['axes'])


class ReduceSum(Operator):
	def __init__(self, axis: int=None, invariant: bool=False):
		self.inputs_count = 1
		self.arguments = {'axis': axis, 'invariant': invariant}

	def f(self, value_a):
		return numpy.sum(value_a, axis=self.arguments['axis'], keepdims=self.arguments['invariant'])

	def bprop(self, engine, symbol_forward, symbol_a):
		forward = engine.bprop(symbol_forward)
		shape_a = engine.shape2(symbol_a)
		axis = self.arguments['axis']
		if axis:
			return [lambda: forward(Expand(axis))(Broadcast(shape_a))]
		else:
			return [lambda: forward(Broadcast(shape_a))]

	def shape(self, shape_a):
		return reduce_shape(shape_a, **self.arguments)


class ReduceMean(Operator):
	def __init__(self, axis: int=None, invariant: bool=False):
		self.inputs_count = 1
		self.arguments = {'axis': axis, 'invariant': invariant}

	def f(self, value_a):
		return numpy.mean(value_a, axis=self.arguments['axis'], keepdims=self.arguments['invariant'])

	def bprop(self, engine, symbol_forward, symbol_a):
		forward = engine.bprop(symbol_forward)
		shape_a = engine.shape2(symbol_a)
		axis = self.arguments['axis']
		if axis:
			return [lambda: forward(Expand(axis))(Broadcast(shape_a))(Div(),shape_a[axis])]
		else:
			return [lambda: forward(Broadcast(shape_a))( Div(),reduce(lambda x, y: x * y, shape_a, 1))]

	def shape(self, shape_a):
		return reduce_shape(shape_a, **self.arguments)


class Expand(Operator):
	def __init__(self, axis: int):
		self.inputs_count = 1
		self.arguments = {'axis': axis}

	def f(self, value_a):
		return numpy.expand_dims(value_a, **self.arguments)

	def bprop(self, engine, symbol_forward, symbol_a):
		forward = engine.bprop(symbol_forward)
		return [lambda: forward(ReduceMean(**self.arguments))]

	def shape(self, shape_a):
		new_shape = list(shape_a)
		new_shape.insert(self.arguments['axis'], 1)
		broadcast_a = [0] * len(shape_a)
		broadcast_a.insert(self.arguments['axis'], 1)
		return tuple(new_shape), tuple(broadcast_a)


class Broadcast(Operator):
	def __init__(self, shape):
		self.inputs_count = 1
		self.arguments = {'shape': shape}

	def f(self, value_a):
		if len(value_a.shape) == 1 and len(self.arguments['shape']) > 1:
			if value_a.shape[0] == self.arguments['shape'][0]:
				value_a = value_a.reshape((value_a.shape[0], 1))
		return numpy.broadcast_to(value_a, **self.arguments)

	def bprop(self, engine, symbol_forward, symbol_a):
		forward = engine.bprop(symbol_forward)
		return [lambda: forward]

	def shape(self, shape_a):
		return element_wise_shape(shape_a, self.arguments['shape'])[:2]


class Pow(Operator):
	def __init__(self):
		self.operator_sign = '**'
		self.inputs_count = 2

	def f(self, value_a, value_b):
		return numpy.power(value_a, value_b)

	def bprop(self, engine, symbol_forward, symbol_a, symbol_b):
		forward = engine.bprop(symbol_forward)
		return [lambda: forward(Mul(),symbol_b)(Mul(),symbol_a(Pow(),symbol_b(Sub(),1))),
				lambda: forward(Mul(),symbol_a(Pow(),symbol_b))(Mul(),symbol_a(Log()))]

	def shape(self, shape_a, shape_b):
		return element_wise_shape(shape_a, shape_b)


class Log(Operator):
	def __init__(self):
		self.inputs_count = 1

	def f(self, value_a):
		return numpy.log(value_a)

	def bprop(self, engine, symbol_forward, symbol_a):
		forward = engine.bprop(symbol_forward)
		return [lambda: forward(Mul(),1)(Div(),symbol_a)]

	def shape(self, shape_a):
		return shape_a, ()


class Where(Operator):
	def __init__(self):
		self.inputs_count = 3

	def f(self, value_condition, value_a, value_b):
		return numpy.array(numpy.where(value_condition, value_a, value_b), dtype=float)

	def bprop(self, engine, symbol_forward, symbol_condition, symbol_a, symbol_b):
		forward = engine.bprop(symbol_forward)
		
		return [lambda: constant(0),
				lambda: forward(Mul(),V(oprator=Where(),inputs=as_symbols([symbol_condition, forward, 0]))),
				lambda: forward(Mul(),V(oprator=Where(),inputs=as_symbols([symbol_condition, 0,forward])))]

	def shape(self, shape_condition, shape_a, shape_b):
		return element_wise_shape(shape_condition, shape_a, shape_b)


class Eq(Operator):
	def __init__(self):
		self.operator_sign = '=='
		self.inputs_count = 2

	def f(self, value_a, value_b):
		return numpy.equal(value_a, value_b)

	def bprop(self, engine, symbol_forward, symbol_a, symbol_b):
		forward = engine.bprop(symbol_forward)
		return [lambda: V(operator=Where(),inputs=as_symbols([symbol_a(Eq(),symbol_b), forward, 0])),
				lambda: V(operator=Where(),inputs=as_symbols([symbol_a(Eq(),symbol_b), forward, 0]))]

	def shape(self, shape_a, shape_b):
		return element_wise_shape(shape_a, shape_b)


class NEq(Operator):
	def __init__(self):
		self.operator_sign = '!='
		self.inputs_count = 2

	def f(self, value_a, value_b):
		return numpy.not_equal(value_a, value_b)

	def bprop(self, engine, symbol_forward, symbol_a, symbol_b):
		forward = engine.bprop(symbol_forward)
		return [lambda:V(operator=Where(),inputs=as_symbols([symbol_a(NEq(),symbol_b), forward,0])),
				lambda:V(operator=Where(),inputs=as_symbols([symbol_a(NEq(),symbol_b), forward,0])),]

	def shape(self, shape_a, shape_b):
		return element_wise_shape(shape_a, shape_b)


class Lt(Operator):
	def __init__(self):
		self.operator_sign = '<'
		self.inputs_count = 2

	def f(self, value_a, value_b):
		return numpy.less(value_a, value_b)

	def bprop(self, engine, symbol_forward, symbol_a, symbol_b):
		forward = engine.bprop(symbol_forward)
		return [lambda:V(operator=Where(),inputs=as_symbols([symbol_a(Lt(),symbol_b), forward, 0])),
				lambda:V(operator=Where(),inputs=as_symbols([symbol_a(Lt(),symbol_b), forward, 0])),]

	def shape(self, shape_a, shape_b):
		return element_wise_shape(shape_a, shape_b)


class LEq(Operator):
	def __init__(self):
		self.operator_sign = '<='
		self.inputs_count = 2

	def f(self, value_a, value_b):
		return numpy.less_equal(value_a, value_b)

	def bprop(self, engine, symbol_forward, symbol_a, symbol_b):
		forward = engine.bprop(symbol_forward)
		return [lambda:V(operator=Where(),inputs=as_symbols([symbol_a(LEq(),symbol_b), forward, 0])),
				lambda:V(operator=Where(),inputs=as_symbols([symbol_a(LEq(),symbol_b), forward, 0])),]

	def shape(self, shape_a, shape_b):
		return element_wise_shape(shape_a, shape_b)


class Gt(Operator):
	def __init__(self):
		self.operator_sign = '>'
		self.inputs_count = 2

	def f(self, value_a, value_b):
		return numpy.greater(value_a, value_b)

	def bprop(self, engine, symbol_forward, symbol_a, symbol_b):
		forward = engine.bprop(symbol_forward)
		return [lambda:V(operator=Where(),inputs=as_symbols([symbol_a(Gt(),symbol_b), forward, 0])),
				lambda:V(operator=Where(),inputs=as_symbols([symbol_a(Gt(),symbol_b), forward, 0])),]

	def shape(self, shape_a, shape_b):
		return element_wise_shape(shape_a, shape_b)


class GEq(Operator):
	def __init__(self):
		self.operator_sign = '>='
		self.inputs_count = 2

	def f(self, value_a, value_b):
		return numpy.greater_equal(value_a, value_b)

	def bprop(self, engine, symbol_forward, symbol_a, symbol_b):
		forward = engine.bprop(symbol_forward)
		return [lambda:V(operator=Where(),inputs=as_symbols([symbol_a(GEq(),symbol_b), forward, 0])),
				lambda:V(operator=Where(),inputs=as_symbols([symbol_a(GEq(),symbol_b), forward, 0])),]

	def shape(self, shape_a, shape_b):
		return element_wise_shape(shape_a, shape_b)


class Max(Operator):
	def __init__(self):
		self.inputs_count = 2

	def f(self, value_a, value_b):
		return numpy.maximum(value_a, value_b)

	def bprop(self, engine, symbol_forward, symbol_a, symbol_b):
		forward = engine.bprop(symbol_forward)
		return [lambda:V(operator=Where(),inputs=as_symbols([symbol_a(Gt(),symbol_b), forward, 0])),
				lambda:V(operator=Where(),inputs=as_symbols([symbol_a(Lt(),symbol_b), forward, 0])),]

	def shape(self, shape_a, shape_b):
		return element_wise_shape(shape_a, shape_b)


class Min(Operator):
	def __init__(self):
		self.inputs_count = 2

	def f(self, value_a, value_b):
		return numpy.minimum(value_a, value_b)

	def bprop(self, engine, symbol_forward, symbol_a, symbol_b):
		forward = engine.bprop(symbol_forward)
		return [lambda:V(operator=Where(),inputs=as_symbols([symbol_a(Lt(),symbol_b), forward, 0])),
				lambda:V(operator=Where(),inputs=as_symbols([symbol_a(Gt(),symbol_b), forward, 0])),]

	def shape(self, shape_a, shape_b):
		return element_wise_shape(shape_a, shape_b)


class Sin(Operator):
	def __init__(self):
		self.inputs_count = 1

	def f(self, value_a):
		return numpy.sin(value_a)

	def bprop(self, engine, symbol_forward, symbol_a):
		forward = engine.bprop(symbol_forward)
		return [lambda: forward(Mul(),symbol_a(Cos()))]

	def shape(self, shape_a):
		return shape_a, ()


class Cos(Operator):
	def __init__(self):
		self.inputs_count = 1

	def f(self, value_a):
		return numpy.cos(value_a)

	def bprop(self, engine, symbol_forward, symbol_a):
		forward = engine.bprop(symbol_forward)
		return [lambda: forward(Mul(),symbol_a(Sin())(Neg()))]

	def shape(self, shape_a):
		return shape_a, ()


class Tan(Operator):
	def __init__(self):
		self.inputs_count = 1

	def f(self, value_a):
		return numpy.tan(value_a)

	def bprop(self, engine, symbol_forward, symbol_a):
		forward = engine.bprop(symbol_forward)
		return [lambda: forward(Mul(),1)(Div(),symbol_a(Cos())(Pow(),2))]

	def shape(self, shape_a):
		return shape_a, ()


class ArcSin(Operator):
	def __init__(self):
		self.inputs_count = 1

	def f(self, value_a):
		return numpy.arcsin(value_a)

	def bprop(self, engine, symbol_forward, symbol_a):
		forward = engine.bprop(symbol_forward)
		return [lambda: forward(Mul(),1)(Div(),constant(1)(Sub(),symbol_a(Pow(),2))(Pow(),0.5))]

	def shape(self, shape_a):
		return shape_a, ()


class ArcCos(Operator):
	def __init__(self):
		self.inputs_count = 1

	def f(self, value_a):
		return numpy.arccos(value_a)

	def bprop(self, engine, symbol_forward, symbol_a):
		forward = engine.bprop(symbol_forward)
		return [lambda: forward(Mul(),-1)(Div(),constant(1)(Sub(),symbol_a(Pow(),constant(2)))(Pow(),0.5))]

	def shape(self, shape_a):
		return shape_a, ()


class ArcTan(Operator):
	def __init__(self):
		self.inputs_count = 1

	def f(self, value_a):
		return numpy.arctan(value_a)

	def bprop(self, engine, symbol_forward, symbol_a):
		forward = engine.bprop(symbol_forward)
		return [lambda: forward(Mul(),1)(Div(),constant(1)(Plus(),symbol_a(Pow(),2)))]

	def shape(self, shape_a):
		return shape_a, ()


class Sinh(Operator):
	def __init__(self):
		self.inputs_count = 1

	def f(self, value_a):
		return numpy.sinh(value_a)

	def bprop(self, engine, symbol_forward, symbol_a):
		forward = engine.bprop(symbol_forward)
		return [lambda: forward(Mul(),symbol_a(Cosh()))]

	def shape(self, shape_a):
		return shape_a, ()


class Cosh(Operator):
	def __init__(self):
		self.inputs_count = 1

	def f(self, value_a):
		return numpy.cosh(value_a)

	def bprop(self, engine, symbol_forward, symbol_a):
		forward = engine.bprop(symbol_forward)
		return [lambda: forward(Mul(),symbol_a(Sinh()))]

	def shape(self, shape_a):
		return shape_a, ()


class Tanh(Operator):
	def __init__(self):
		self.inputs_count = 1

	def f(self, value_a):
		return numpy.tanh(value_a)

	def bprop(self, engine, symbol_forward, symbol_a):
		forward = engine.bprop(symbol_forward)
		return [lambda: forward(Mul(),constant(1)(Sub(),symbol_a(Tanh())(Pow(),2)))]

	def shape(self, shape_a):
		return shape_a, ()


class ArcSinh(Operator):
	def __init__(self):
		self.inputs_count = 1

	def f(self, value_a):
		return numpy.arcsinh(value_a)

	def bprop(self, engine, symbol_forward, symbol_a):
		forward = engine.bprop(symbol_forward)
		return [lambda: forward(Mul(),1)(Div(),symbol_a(Pow(),2)(Plus(),1)(Pow(),0.5))]

	def shape(self, shape_a):
		return shape_a, ()


class ArcCosh(Operator):
	def __init__(self):
		self.inputs_count = 1

	def f(self, value_a):
		return numpy.arccosh(value_a)

	def bprop(self, engine, symbol_forward, symbol_a):
		forward = engine.bprop(symbol_forward)
		return [lambda: forward(Mul(),1)(Div(),symbol_a(Pow(),2)(Sub(),1)(Pow(),0.5))]

	def shape(self, shape_a):
		return shape_a, ()


class ArcTanh(Operator):
	def __init__(self):
		self.inputs_count = 1

	def f(self, value_a):
		return numpy.arctanh(value_a)

	def bprop(self, engine, symbol_forward, symbol_a):
		forward = engine.bprop(symbol_forward)
		return [lambda: forward(Mul(),1)(Div(),constant(1)(Sub(),symbol_a(Pow(),2)))]

	def shape(self, shape_a):
		return shape_a, ()


class Exp(Operator):
	def __init__(self):
		self.inputs_count = 1

	def f(self, value_a):
		return numpy.exp(value_a)

	def bprop(self, engine, symbol_forward, symbol_a):
		forward = engine.bprop(symbol_forward)
		return [lambda: forward(Mul(),symbol_a(Exp()))]

	def shape(self, shape_a):
		return shape_a, ()


class SliceAssign(Operator):
	def __init__(self, slice_tuple):
		self.inputs_count = 2
		self.arguments = {'slice_tuple': slice_tuple}

	def f(self, value_a, value_b):
		value_a[self.arguments['slice_tuple']] = value_b
		return value_a

	def bprop(self, engine, symbol_forward, symbol_a, symbol_b):
		slice_tuple = self.arguments['slice_tuple']
		forward = engine.bprop(symbol_forward)
		return [lambda: forward(SliceAssign(slice_tuple),numpy.zeros(engine.shape2(symbol_b))),
				lambda: forward[slice_tuple]]

	def shape(self, shape_a, shape_b):
		shape_select = slice_shape(shape_a, self.arguments['slice_tuple'])[0]
		shape_package = element_wise_shape(shape_select, shape_b)
		if shape_package[0] != shape_select:
			raise ValueError('Can not assign: {} to {} with {}'.format(shape_b, shape_a, self.arguments['slice_tuple']))
		return shape_a, (), shape_package[2]


class SliceSelect(Operator):
	def __init__(self, slice_tuple):
		self.inputs_count = 1
		self.arguments = {'slice_tuple': slice_tuple}

	def f(self, value_a):
		return value_a[self.arguments['slice_tuple']]

	def bprop(self, engine, symbol_forward, symbol_a):
		forward = engine.bprop(symbol_forward)
		symbol_zero = constant(numpy.zeros(engine.shape2(symbol_a)))
		return [lambda: symbol_zero(SliceAssign(forward), self.arguments['slice_tuple'])]

	def shape(self, shape_a):
		return slice_shape(shape_a, self.arguments['slice_tuple'])


class Concatenate(Operator):
	def __init__(self, axis: int=0):
		self.inputs_count = 2
		self.arguments = {'axis': axis}

	def f(self, value_a, value_b):
		return numpy.concatenate((value_a, value_b), **self.arguments)

	def bprop(self, engine, symbol_forward, symbol_a, symbol_b):
		forward = engine.bprop(symbol_forward)
		split_dimension = engine.shape2(symbol_a)[self.arguments['axis']]
		total_dimension = engine.shape2(symbol_forward)[self.arguments['axis']]
		return [lambda: forward[[slice(None)] * self.arguments['axis'] + [slice(0, split_dimension)]],
				lambda: forward[[slice(None)] * self.arguments['axis'] + [slice(split_dimension, total_dimension)]]]

	def shape(self, *shapes):
		return concatenate_shape(self.arguments['axis'], *shapes)


class Rotate90(Operator):
	def __init__(self, count: int=1, axes: tuple=None):
		self.inputs_count = 1
		self.arguments = {'count': count, 'axes': axes}

	def f(self, value_a):
		return numpy.rot90(value_a, k=self.arguments['count'], axes=self.arguments['axes'])

	def bprop(self, engine, symbol_forward, symbol_a, symbol_b):
		forward = engine.bprop(symbol_forward)
		return [lambda: forward(Rotate90(-self.arguments['count'] & 3, self.arguments['axes']))]

	def shape(self, shape_a):
		return rotate90_shape(shape_a, **self.arguments)


class Flip(Operator):
	def __init__(self, axis: int):
		self.inputs_count = 1
		self.arguments = {'axis': axis}

	def f(self, value_a):
		return numpy.flip(value_a, self.arguments['axis'])

	def bprop(self, engine, symbol_forward, symbol_a):
		forward = engine.bprop(symbol_forward)
		return [lambda: forward(Flip(self.arguments['axis']))]

	def shape(self, shape_a):
		return shape_a, ()


class Reshape(Operator):
	def __init__(self, shape):
		self.inputs_count = 1
		self.arguments = {'shape': shape}

	def f(self, value_a):
		return numpy.reshape(value_a, self.arguments['shape'])

	def bprop(self, engine, symbol_forward, symbol_a):
		forward = engine.bprop(symbol_forward)
		return [lambda: forward(Reshape(engine.shape2(symbol_a)))]

	def shape(self, shape_a):
		return self.arguments['shape'], ()


class Spread(Operator):
	def __init__(self, position):
		self.inputs_count = 1
		self.arguments = {'position': position}

	def f(self, value_a):
		shape_a = value_a.shape
		spread_dimension = numpy.prod(shape_a[self.arguments['position']:])#reduce(lambda a, b: a * b, shape_a[self.arguments['position']:])
		new_shape = shape_a[:self.arguments['position']] + (spread_dimension,)
		return numpy.reshape(value_a, new_shape)

	def bprop(self, engine, symbol_forward, symbol_a):
		forward = engine.bprop(symbol_forward)
		return [lambda: forward(Reshape(engine.shape2(symbol_a)))]

	def shape(self, shape_a):
		spread_dimension = reduce(lambda a, b: a * b, shape_a[self.arguments['position']:])
		new_shape = shape_a[:self.arguments['position']] + (spread_dimension,)
		return new_shape, ()


# def negative(a):
#     return V(operator=Neg(), inputs=as_symbols([a]))
#  
#  
# def absolute(a):
#     return V(operator=Abs(), inputs=as_symbols([a]))
#  
#  
# def plus(a, b):
#     return V(operator=Plus(), inputs=as_symbols([a, b]))
#  
#  
# def subtract(a, b):
#     return V(operator=Sub(), inputs=as_symbols([a, b]))
#  
#  
# def multiply(a, b):
#     return V(operator=Mul(), inputs=as_symbols([a, b]))
#  
#  
# def divide(a, b):
#     return V(operator=Div(), inputs=as_symbols([a, b]))
#  
#  
# def matrix_multiply(a, b):
#     return V(operator=MM(), inputs=as_symbols([a, b]))
#  
#  
# def power(a, b):
#     return V(operator=Pow(), inputs=as_symbols([a, b]))
#  
#  
# def log(a):
#     return V(operator=Log(), inputs=as_symbols([a]))
#  
#  
# def transpose(a, axes: tuple=None):
#     return V(operator=T(axes), inputs=as_symbols([a]))
#  
#  
# def reduce_sum(a, axis: int=None, invariant=False):
#     return V(operator=ReduceSum(axis, invariant), inputs=as_symbols([a]))
#  
#  
# def reduce_mean(a, axis: int=None, invariant=False):
#     return V(operator=ReduceMean(axis, invariant), inputs=as_symbols([a]))
#  
#  
# def expand(a, axis: int):
#     return V(operator=Expand(axis), inputs=as_symbols([a]))
#  
#  
# def broadcast(a, shape):
#     return V(operator=Broadcast(shape), inputs=as_symbols([a]))
#  
#  
# def where(condition, a, b):
#     return V(operator=Where(), inputs=as_symbols([condition, a, b]))
#  
#  
# def equal(a, b):
#     return V(operator=Eq(), inputs=as_symbols([a, b]))
#  
#  
# def not_equal(a, b):
#     return V(operator=NEq(), inputs=as_symbols([a, b]))
#  
#  
# def less(a, b):
#     return V(operator=Lt(), inputs=as_symbols([a, b]))
#  
#  
# def less_equal(a, b):
#     return V(operator=LEq(), inputs=as_symbols([a, b]))
#  
#  
# def greater(a, b):
#     return V(operator=Gt(), inputs=as_symbols([a, b]))
#  
#  
# def greater_equal(a, b):
#     return V(operator=GEq(), inputs=as_symbols([a, b]))
#  
#  
# def maximum(a, b):
#     return V(operator=Max(), inputs=as_symbols([a, b]))
#  
#  
# def minimum(a, b):
#     return V(operator=Min(), inputs=as_symbols([a, b]))
#  
#  
# def sin(a):
#     return V(operator=Sin(), inputs=as_symbols([a]))
#  
#  
# def cos(a):
#     return V(operator=Cos(), inputs=as_symbols([a]))
#  
#  
# def tan(a):
#     return V(operator=Tan(), inputs=as_symbols([a]))
#  
#  
# def arcsin(a):
#     return V(operator=ArcSin(), inputs=as_symbols([a]))
#  
#  
# def arccos(a):
#     return V(operator=ArcCos(), inputs=as_symbols([a]))
#  
#  
# def arctan(a):
#     return V(operator=ArcTan(), inputs=as_symbols([a]))
#  
#  
# def sinh(a):
#     return V(operator=Sinh(), inputs=as_symbols([a]))
#  
#  
# def cosh(a):
#     return V(operator=Cosh(), inputs=as_symbols([a]))
#  
#  
# def tanh(a):
#     return V(operator=Tanh(), inputs=as_symbols([a]))
#  
#  
# def arcsinh(a):
#     return V(operator=ArcSinh(), inputs=as_symbols([a]))
#  
#  
# def arccosh(a):
#     return V(operator=ArcCosh(), inputs=as_symbols([a]))
#  
#  
# def arctanh(a):
#     return V(operator=ArcTanh(), inputs=as_symbols([a]))
#  
#  
# def exp(a):
#     return V(operator=Exp(), inputs=as_symbols([a]))
#  
#  
# def slice_assign(a, b, slice_tuple):
#     return V(operator=SliceAssign(slice_tuple), inputs=as_symbols([a, b]))
#  
#  
# def assign(a, b):
#     return slice_assign(a, b, slice(None))
#  
#  
# def slice_select(a, slice_tuple):
#     return V(operator=SliceSelect(slice_tuple), inputs=as_symbols([a]))
#  
#  
# def concatenate(a, b):
#     return V(operator=Concatenate(), inputs=as_symbols([a, b]))
#  
#  
# def rotate90(a, count, axes):
#     return V(operator=Rotate90(count, axes), inputs=as_symbols([a]))
#  
#  
# def flip(a, axis):
#     return V(operator=Flip(axis), inputs=as_symbols([a]))
#  
# def reshape(a, shape):
#     return V(operator=Reshape(shape), inputs=as_symbols([a]))
#  
#  
# def spread(a, position):
#     return V(operator=Spread(position), inputs=as_symbols([a]))
def where(a,b,c):
	return V(operator=Where(),inputs=as_symbols([a,b,c]))

#algbra---------
class Template:
	active_operator = None
	def __init__(self):
		if self.active_operator is None:
			self.active_sign = None
		else:
			self.active_sign = self.active_operator.__name__

	@staticmethod
	def reduce_symbol(symbol, index: int):
		if not isinstance(symbol,V):raise ValueError("symbol Must be V")
		input_list = symbol.input
		reduce_to_symbol = input_list[index]
		symbol.clear_operator()
		symbol.value = reduce_to_symbol.value
		symbol.name = reduce_to_symbol.name
		if reduce_to_symbol.is_operator():
			symbol.symbolic_compute(reduce_to_symbol.operator, reduce_to_symbol.input)
		else:
			symbol.category = reduce_to_symbol.category

	@staticmethod
	def value_equal(a, b):
		result = a == b
		if isinstance(result, bool) or isinstance(result, numpy.bool_):
			return result
		elif isinstance(result, numpy.ndarray):
			return result.all()
		else:
			raise Exception('Never reached.')

	@staticmethod
	def symbol_equal(a, b):
		if not isinstance(a,V) or not isinstance(b,V):raise ValueError("symbol Must be V")
		return a.symbolic_hash() == b.symbolic_hash()

	def simplify(self, symbol):
		raise ValueError("this is abstract")

class Simplification:
	def __init__(self,templates:list):
		self.__templates = {}
		for template in templates:
			self.register(template())
	def operator_trigger(self, operator: Operator):
		operator_sign = operator.__class__.__name__
		if operator_sign in self.__templates:
			return self.__templates[operator_sign]
		else:
			return set()

	def register(self, template: Template):
		active_operator = template.active_sign
		self.__templates.setdefault(active_operator, set())
		self.__templates[active_operator].add(template)

	def simplify(self, symbol):
		if not isinstance(symbol,V):raise ValueError("symbol Must be V")
		while self.simplify_cycle(symbol):
			pass
	def simplify_cycle(self, symbol):
		if not isinstance(symbol,V):raise ValueError("symbol Must be V")
		effective = False
		templates = list(self.operator_trigger(symbol.operator)) + list(self.__templates[None])
		if templates:
			for template in templates:
				if template.simplify(symbol):
					effective |= True
					break
		for next_symbol in symbol.input:
			if next_symbol.operator is not None:
				effective |= self.simplify_cycle(next_symbol)
		return effective

class TemplateConstant(Template):
	active_operator = None

	def simplify(self, symbol: V):
		if symbol.is_operator():
			for s in symbol.input:
				if not s.is_constant():
					return False
			compute_inputs = [_s.value for _s in symbol.input]
			value = symbol.operator.f(*compute_inputs)
			symbol.clear_operator()
			symbol.value = value
			symbol.category = VCategory.constant
			symbol.name=None
			return True
		else:
			return False


class TemplatePlus(Template):
	active_operator = Plus

	def simplify(self, symbol: V):
		left_symbol, right_symbol = symbol.input
		if left_symbol.is_constant() and self.value_equal(left_symbol.value, 0):
			self.reduce_symbol(symbol, 1)
			return True
		elif right_symbol.is_constant() and self.value_equal(right_symbol.value, 0):
			self.reduce_symbol(symbol, 0)
			return True
		elif self.symbol_equal(left_symbol, right_symbol):
			symbol.clear_operator()
			symbol.symbolic_compute(Mul(), [left_symbol, constant(2)])
			symbol.name=None
			return True
		elif left_symbol.operator!=None: 
			if left_symbol.operator.operator_sign=='-' and left_symbol.operator.inputs_count==1 and self.symbol_equal(left_symbol.input[0],right_symbol):
				symbol.clear_operator()
				symbol.symbolic_compute(Broadcast(()),[constant(0)])
				symbol.name=None
				return True
		elif right_symbol.operator!=None: 
			if right_symbol.operator.operator_sign=='-' and left_symbol.operator.inputs_count==1 and self.symbol_equal(right_symbol.input[0],left_symbol):
				symbol.clear_operator()
				symbol.symbolic_compute(Broadcast(()),[constant(0)])
				symbol.name=None
				return True
		else:
			return False


class TemplateSubtract(Template):
	active_operator = Sub

	def simplify(self, symbol: V):
		left_symbol, right_symbol = symbol.input
		if right_symbol.is_constant() and self.value_equal(left_symbol.value, 0):
			symbol.clear_operator()
			symbol.clear_input()
			symbol.symbolic_compute(Neg(), [right_symbol])
			symbol.name=None
			return True
		elif right_symbol.is_constant() and self.value_equal(right_symbol.value, 0):
			self.reduce_symbol(symbol, 0)
			return True
		elif self.symbol_equal(left_symbol, right_symbol):
			symbol.clear_operator()
			symbol.symbolic_compute(Broadcast(()), [left_symbol, constant(0)])
			symbol.name=None
			return True
		else:
			return False


class TemplateDivide(Template):
	active_operator = Div

	def simplify(self, symbol: V):
		left_symbol, right_symbol = symbol.input
		if right_symbol.is_constant() and self.value_equal(right_symbol.value, 1):
			self.reduce_symbol(symbol, 0)
			return True
		elif left_symbol.is_constant() and self.value_equal(left_symbol.value, 0):
			symbol.clear_operator()
			symbol.symbolic_compute(SliceAssign(slice(None)), [left_symbol, constant(0)])
			symbol.name=None
			return True
		elif self.symbol_equal(left_symbol, right_symbol):
			symbol.clear_operator()
			symbol.symbolic_compute(SliceAssign(slice(None)), [left_symbol, constant(1)])
			symbol.name=None
			return True
		else:
			return False


class TemplateMultiply(Template):
	active_operator = Mul

	def simplify(self, symbol: V):
		left_symbol, right_symbol = symbol.input
		if left_symbol.is_constant() and self.value_equal(left_symbol.value, 1):
			self.reduce_symbol(symbol, 1)
			return True
		elif right_symbol.is_constant() and self.value_equal(right_symbol.value, 1):
			self.reduce_symbol(symbol, 0)
			return True
		elif (left_symbol.is_constant() and self.value_equal(left_symbol.value, 0)) or(right_symbol.is_constant() and self.value_equal(right_symbol.value, 0)) :
			symbol.clear_operator()
			symbol.symbolic_compute(Broadcast(()),[constant(0)])
			symbol.name=None
			return True
		elif self.symbol_equal(left_symbol, right_symbol):
			symbol.clear_operator()
			symbol.symbolic_compute(Pow(), [left_symbol, constant(2)])
			symbol.name=None
			return True
		elif left_symbol.operator!=None: 
			if left_symbol.operator.operator_sign=='/' and left_symbol.input[0].is_constant() and self.symbol_equal(left_symbol.input[1],right_symbol):
				if left_symbol.input[0].value==1:
					symbol.clear_operator()
					symbol.symbolic_compute(Broadcast(()),[constant(1)])
					symbol.name=None
					return True
		elif right_symbol.operator!=None: 
			if right_symbol.operator.operator_sign=='/' and right_symbol.input[0].is_constant() and self.symbol_equal(right_symbol.input[1],left_symbol):
				if right_symbol.input[0].value==1:
					symbol.clear_operator()
					symbol.symbolic_compute(Broadcast(()),[constant(1)])
					symbol.name=None
					return True
		else:
			return False


class TemplatePower(Template):
	active_operator = Pow
	def simplify(self, symbol: V):
		left_symbol, right_symbol = symbol.input
		if (self.value_equal(left_symbol.value, 1) and left_symbol.is_constant() and right_symbol.is_constant()) or (self.value_equal(right_symbol.value, 1) and right_symbol.is_constant()):
			self.reduce_symbol(symbol, 0)
			return True
		elif self.value_equal(left_symbol.value, 1) and left_symbol.is_constant() :
			symbol.clear_operator()
			symbol.symbolic_compute(Broadcast(right_symbol.shape),[constant(1)])    
			symbol.name=None
			return True
		elif self.value_equal(right_symbol.value, 0) and right_symbol.is_constant():
			symbol.clear_operator()
			symbol.symbolic_compute(Broadcast(()),[constant(1)])    
			symbol.name=None
			return True
		else:
			return False

class TemplateMM(Template):
	active_operator = MM
	maxlen=4;
	def simplify(self, symbol:V):
		ls=g(symbol)
		if len(ls)>self.maxlen:
			self.maxlen=len(ls)
#             print(ls)
			symbol2=rebuildChain(ls)            
			symbol.clear_operator()
			symbol.symbolic_compute(SliceSelect(slice(None,None,None)),[symbol2])    
			symbol.name=None
			return False


def chainTrace(i,j,s):
	if i==j:
		return []
	ls=[]
	ls+=(chainTrace(i,s[(i,j)],s))
	ls+=(chainTrace(s[(i,j)]+1,j,s))
	ls.append(s[(i,j)])
	return ls    
def matChain(p:list):
	n=len(p)-1
	m,s={},{}
	for i in range(1,n+1):m[(i,i)]=0
	for r in range(2,n+1):
		for i in range(1,n-r+2):
			j=i+r-1
			m[(i,j)]=m[(i+1,j)]+p[i-1]*p[i]*p[j]
			s[(i,j)]=i;
			for k in range(i+1,j):
				t=m[(i,k)]+m[(k+1,j)]+p[i-1]*p[k]*p[j]
				if t<m[(i,j)]:
					m[(i,j)]=t;s[(i,j)]=k    
	return chainTrace(1, n, s)
def rebuildChain(ls):
	if len(ls)<3:return
	ls1=[ls[0].shape[0]]+[i.shape[1] for i in ls]
	m=matChain(ls1)
	nls=ls[m[0]-1](MM(),ls[m[0]])
	ls[m[0]-1]=nls
	ls[m[0]]=None
	for i in m[1:]:
		j=0
		k=1
		while ls[i+j]==None:j+=1
		while ls[i-k]==None:k+=1
		nls=ls[i-k](MM(),ls[i+j])
		ls[i-k]=nls
		ls[i+j]=None
	return nls

def g(l):
	if l.operator==None or l.operator.operator_sign!='@':
		return [l]
	ls =[]
	for i in range(l.operator.inputs_count):
		ls+=g(l.input[i])
	return ls

default_templates = [
	TemplateConstant,
	TemplatePlus,
	TemplateSubtract,
	TemplateMultiply,
	TemplateDivide,
	TemplatePower,
	TemplateMM,
]
simp=Simplification(default_templates)





import numpy


def generate_label_matrix(classification, default: int=0):
    index_map = {}
    index_reverse_map = {}
    index = 0
    for c in classification:
        if c not in index_map:
            index_map[c] = index
            index_reverse_map[index] = c
            index += 1
    dimension = len(index_map)
    class_matrix = default * numpy.ones([len(classification), dimension])
    for i, c in enumerate(classification):
        class_matrix[i, index_map[c]] = 1
    return class_matrix, index_map, index_reverse_map



import trik
A = trik.constant([[1, 2], [1, 3]], name='A')
x = trik.V([[1, 0,1],[1,0,1]], name='x')
x1 =trik.V([[1, 0],[1,1],[1,0]], name='x1')
b = trik.constant([3, 4], name='b')
#~ loss = pd.reduce_sum(((A @ x@(1**x1) + b)) ** 2) / 2
loss=A(trik.MM(),x)(trik.MM(),trik.constant(1)(trik.Pow(),x1))(trik.Plus(),b)(trik.Pow(),2)(trik.ReduceSum())(trik.Div(),2)
 
print('loss formula =\n{}\n'.format(loss))
print('loss =\n{}\n'.format(loss.val()))
  
x_gradient = loss.bprop(x)
print('x grad formula =\n{}\n'.format(x_gradient))
print('x grad =\n{}\n'.format(x_gradient.val()))


import pickle


def save(file: str, thing):
    with open(file, 'wb') as save_file:
        pickle.dump(thing, save_file)


def load(file: str):
    with open(file, 'rb') as load_file:
        return pickle.load(load_file)



# -*- coding:utf-8 -*-
import numpy as np
import matplotlib.pyplot as plt
import trik

# 每类随机生成点的个数。
points_sum = 100

# 产生一个互相环绕的螺旋形数据分布。
c1_x, c1_y, c2_x, c2_y = [], [], [], []
r_step = 5 / points_sum
theta_step = 3 * np.pi / points_sum
r = 0
theta = 0
for _ in range(points_sum):
    c1_x.append(r * np.cos(theta))
    c1_y.append(r * np.sin(theta))
    c2_x.append(-r * np.cos(theta))
    c2_y.append(-r * np.sin(theta))
    r += r_step
    theta += theta_step
c_x = c1_x + c2_x
c_y = c1_y + c2_y

# 定义每个点的分类类别。
classification = trik.yu.generate_label_matrix([0] * points_sum + [1] * points_sum)[0]

# 定义符号。
A = trik.V(np.array([c_x, c_y]).transpose(), name='A')
W1 = trik.V(np.random.random((2, 8)), name='W1')  # 输入层到隐含层的权重矩阵。
W2 = trik.V(np.random.random((8, 8)), name='W2')  # 第1层隐含层到输出层的权重矩阵。
W3 = trik.V(np.random.random((8, 2)), name='W3')  # 第2层隐含层到输出层的权重矩阵。
B1 = trik.V(np.random.random((1, 8)), name='B1')  # 第1层隐含层的偏置。
B2 = trik.V(np.random.random((1, 8)), name='B2')  # 第2层隐含层的偏置。
B3 = trik.V(np.random.random((1, 2)), name='B3')  # 输出层的偏置。

# 构建2x8x8x2网络，使用ReLu激活函数。
model = trik.relu(trik.relu(trik.relu(A(trik.MM(),W1)(trik.Plus(),B1))(trik.MM(), W2)(trik.Plus(),B2))(trik.MM(), W3)(trik.Plus(),B3))

# 使用Softmax loss。
loss = trik.softmax_loss(model, trik.constant(classification))

print(loss)

loss.variables=[W1, W2, W3, B1, B2, B3]
optimizer = trik.SGD(0.1,0,0)
for epoch in range(10000):
    optimizer.minimize(loss)
    if epoch % 100 == 0:  # 每100次epoch检查一次loss。
        loss_value = loss.val()
        print('{},loss = {:.8f}'.format(epoch,loss_value))
        if loss_value < 0.01:  # loss阈值。
            break

predict = trik.where(trik.constant([[-1, 1]])(trik.Mul(), model)(trik.ReduceMean(axis=1))(trik.Lt(),0), -1, 1)

h = 0.1
# 生成预测采样点网格。
x, y = np.meshgrid(np.arange(np.min(c_x) - 1, np.max(c_x) + 1, h), np.arange(np.min(c_y) - 1, np.max(c_y) + 1, h))
predict.bind = {A: np.array([x.ravel(), y.ravel()]).transpose()}

z = predict.val().reshape(x.shape)

# 绘制图像。
plt.title('Paradox implement 2x8x8x2 Neural Network')
plt.plot(c1_x, c1_y, 'ro', label='Category 1')
plt.plot(c2_x, c2_y, 'bo', label='Category 2')
plt.contourf(x, y, z, 2, cmap='RdBu', alpha=.6)
plt.legend()
plt.show()



from .sheng import *
from .ya import *
from .yu import *
from numpy import array,arange,meshgrid,pi,cos,sin,exp,log,tan,tanh,sinh,cosh,arctan2,arcsin,arccos,arctanh,arcsinh,arccosh,linalg,fft,polynomial,random,matlib


# -*- coding:utf-8 -*-
import numpy as np
import numpy as np
import matplotlib.pyplot as plt
import trik

# 每类随机生成点的个数。
points_sum = 50

# 调用paradox的数据生成器生成3x3网格状数据。
data = trik.grid_data(points_sum, raw=3, column=3)

# 组合数据。
c_x = data[0][0] + data[1][0]
c_y = data[0][1] + data[1][1]

# 定义每个点的分类类别。
classification = trik.generate_label_matrix([0] * len(data[0][0]) + [1] * len(data[1][0]))[0]

# 调用高层API生成2x8x8x8x8x2的网络，5层网络。
model = trik.Model()
model.add(trik.Dense(8, in_dim=2))  # 2维输入8维输出的全连接层。
model.add(trik.Activation('tanh'))  # 使用tanh激活函数。
model.add(trik.Dense(8))
model.add(trik.Activation('tanh'))
model.add(trik.Dense(8))
model.add(trik.Activation('tanh'))
model.add(trik.Dense(8))
model.add(trik.Activation('tanh'))
model.add(trik.Dense(2))
model.add(trik.Activation('tanh'))
model.loss('softmax')  # 使用softmax loss。

# 使用梯度下降优化器。
model.optimizer('sgd')

# 执行训练。
model.train(np.array([c_x, c_y]).transpose(), classification, epochs=2000)

# 设置网格密度为0.1。
h = 0.1

# 生成预测采样点网格。
x, y = np.meshgrid(np.arange(np.min(c_x) - .1, np.max(c_x) + .1, h), np.arange(np.min(c_y) - .1, np.max(c_y) + .1, h))

# 生成采样点预测值。
z = model.predict(np.array([x.ravel(), y.ravel()]).transpose()).argmax(axis=1).reshape(x.shape)

# 绘制图像。
plt.title('2x8x8x8x8x2 Grid Classification')
plt.plot(data[0][0], data[0][1], 'ro', label='Category 1')
plt.plot(data[1][0], data[1][1], 'bo', label='Category 2')
plt.contourf(x, y, z, 2, cmap='RdBu', alpha=.6)
plt.legend()
plt.show()



# -*- coding:utf-8 -*-
import numpy as np
import matplotlib.pyplot as plt
import trik

# sin函数采样点数。
points_number = 100

# 生成训练数据。
c_x = np.arange(points_number).reshape([points_number, 1])
c_y = np.sin(c_x / 5)

# 构造1x32x1回归神经网络（最后一层不激活）。
model = trik.Model()
model.add(trik.Dense(16, in_dim=1))  # 1维输入32维输出的全连接层。
model.add(trik.Activation('tanh'))  # 使用tanh激活函数。
model.add(trik.Dense(16))  # 1维输入32维输出的全连接层。
model.add(trik.Activation('tanh'))  # 使用tanh激活函数。
model.add(trik.Dense(1))
model.loss('mse')

# 使用Adam下降优化器。
model.optimizer('adam',0.01)

# 执行训练。
model.train(c_x, c_y, epochs=2000)

# 生成预测数据。
x = np.arange(-5, 135, 0.1)
y = model.predict(x.reshape([x.shape[0], 1]))

# 绘制图像。
plt.title('1x64x1 Regression Neural Network')
plt.plot(c_x, c_y, 'ro', label='Sin(x)')
plt.plot(x,y,'b', label='Regression')
plt.legend()
plt.show()



import time
from ..sheng import V,numpy
from .connection import ConnectionLayer, BatchNormalization, Connection
from .cnnlayer import ConvolutionLayer, Convolution


class Plugin:
    network = None
    enable = True

    def bind_network(self, network):
        self.network = network

    def begin_training(self):
        pass

    def end_training(self):
        pass

    def begin_epoch(self):
        pass

    def end_epoch(self):
        pass

    def begin_iteration(self):
        pass

    def end_iteration(self):
        pass

    def begin_predict(self):
        pass

    def end_predict(self):
        pass


class TrainingStatePlugin(Plugin):
    def __init__(self, state_cycle: int=100, auto_cycle: bool=True):
        self.state_cycle = state_cycle
        self.start_time = None
        self.cycle_start_time = None
        self.iteration_start = 0
        self.count = 0
        self.average_speed = 0
        self.auto_cycle = auto_cycle
        self.cycle = 1

    def begin_training(self):
        self.start_time = time.time()
        self.cycle_start_time = self.start_time
        self.count = 0
        self.average_speed = 0
        self.cycle = 1

    def end_training(self):
        print('\n[Training Complete]\nDuration: {}\nAverage Speed: {:.2f}(iterations/s)'.format(
            time.strftime('%H:%M:%S', time.gmtime(time.time() - self.start_time)),
            self.average_speed
        ))

    def begin_iteration(self):
        if self.network.iteration % self.state_cycle == 0:
            self.cycle_start_time = time.time()
            self.iteration_start = self.network.iteration

    def end_iteration(self):
        if (self.network.iteration % self.state_cycle == 0 and self.network.iteration // self.state_cycle == self.cycle) or self.network.iterations == self.network.iteration:
            speed = (self.network.iteration - self.iteration_start + 1) / (time.time() - self.cycle_start_time)
            self.average_speed = self.average_speed * (self.count / (self.count + 1)) + speed / (self.count + 1)
            if self.auto_cycle:
                cycle = int(self.average_speed)
                self.state_cycle = cycle if cycle else 1
            self.count += 1
            self.cycle = int(self.network.iteration / self.state_cycle + 1.5)
            loss_value = self.network.engine.val()
            print('Training State [epoch = {}/{}, loss = {:.8f}, speed = {:.2f}(iterations/s)]{}'.format(
                self.network.epoch,
                self.network.epochs,
                loss_value,
                speed,
                time.time() - self.cycle_start_time
            ))


class VariableMonitorPlugin(Plugin):
    def __init__(self, layer_name=None, for_iteration: bool=True):
        self.__layer_name = None
        self.__for_iteration = for_iteration
        self.set_layer_name(layer_name)

    def get_layer_name(self):
        return self.__layer_name

    def set_layer_name(self, layer_name):
        if isinstance(layer_name, str):
            self.__layer_name = [layer_name]
        else:
            self.__layer_name = layer_name

    layer_name = property(get_layer_name, set_layer_name)

    def output_variables(self):
        if self.__layer_name is None:
            self.__layer_name = list(self.network.layer_name_map())
        for layer_name in self.__layer_name:
            layer = self.network.get_layer(layer_name)
            if isinstance(layer, ConnectionLayer) or isinstance(layer, Connection):
                layer = layer.connection_layer()
                weight, bias = layer.variables()
                print('[{}]: Weight = \n{}'.format(layer_name, weight.value))
                print('[{}]: Bias = \n{}'.format(layer_name, bias.value))
            elif isinstance(layer, ConvolutionLayer) or isinstance(layer, Convolution):
                layer = layer.convolution_layer()
                kernel = layer.kernel()
                print('[{}]: Kernel = \n{}'.format(layer_name, kernel.value))

    def end_iteration(self):
        if self.__for_iteration:
            self.output_variables()

    def end_epoch(self):
        if not self.__for_iteration:
            self.output_variables()


class BatchNormalizationPlugin(Plugin):
    def __init__(self):
        self.__batch_normalization_layer = []

    def begin_training(self):
        for layer in self.network.layer_stack():
            if isinstance(layer, BatchNormalization):
                self.__batch_normalization_layer.append([layer, [], []])

    def end_epoch(self):
        for layer_tuple in self.__batch_normalization_layer:
            layer_tuple[1] = [numpy.mean(numpy.array(layer_tuple[1]), axis=0)]
            layer_tuple[2] = [numpy.mean(numpy.array(layer_tuple[2]), axis=0)]

    def end_iteration(self):
        for layer_tuple in self.__batch_normalization_layer:
            layer_mean_symbol, layer_variance_symbol = layer_tuple[0].normalization_symbol()
            normalization_engine = V();normalization_engine.set_variables(layer_variance_symbol)
            normalization_engine.bind = self.network.engine.bind
            layer_tuple[2].append(normalization_engine.val())
            layer_tuple[1].append(normalization_engine.value_cache[layer_mean_symbol])

    def begin_predict(self):
        value_cache = {}
        for layer_tuple in self.__batch_normalization_layer:
            layer_mean_symbol, layer_variance_symbol = layer_tuple[0].normalization_symbol()
            value_cache[layer_mean_symbol] = numpy.mean(numpy.array(layer_tuple[1]), axis=0)
            value_cache[layer_variance_symbol] = numpy.mean(numpy.array(layer_tuple[2]), axis=0)
        self.network.predict_engine.value_cache = value_cache


default_plugin = [
    ('Training State', TrainingStatePlugin(), True),
    ('Variable Monitor', VariableMonitorPlugin(), False),
    ('Batch Normalization', BatchNormalizationPlugin(), True),
]



from ..sheng import V, variable,numpy
from . import ConnectionLayer
from ..sheng.conv import \
    convolution_nd, \
    max_pooling_nd, \
    average_pooling_nd, \
    average_unpooling_nd, \
    convolution_nd_shape, \
    pooling_nd_shape, \
    unpooling_nd_shape


class ConvolutionLayer(ConnectionLayer):
    kernel_shape = None
    mode = None
    input_shape = None

    def __init__(self, kernel_shape, mode, input_shape=None):
        ConnectionLayer.__init__(self)
        self.kernel_shape = kernel_shape
        self.mode = mode
        self.input_shape = input_shape
        self._kernel = variable(numpy.random.normal(0, 1, self.kernel_shape))

    def connection(self, input_symbol: V):
        pass

    def variables(self):
        pass

    def convolution_layer(self):
        return self

    def kernel(self):
        return self._kernel


class ConvolutionNDLayer(ConvolutionLayer):
    def __init__(self, kernel_shape: tuple, mode, dimension: int, input_shape: tuple=None):
        ConvolutionLayer.__init__(self, kernel_shape, mode, input_shape)
        self.__dimension = dimension
        self.__output_symbol = None

    def get_output_shape(self):
        return convolution_nd_shape(self.input_shape, self.kernel_shape, self.__dimension, self.mode)[0]

    output_shape = property(get_output_shape, ConnectionLayer.set_output_shape)

    def connection(self, input_symbol: V):
        if self.__output_symbol is None:
            self.__output_symbol = convolution_nd(input_symbol, self._kernel, self.__dimension, self.mode)
        return self.__output_symbol

    def variables(self):
        return [self.kernel()]

    def weights(self):
        return [self.kernel()]


class Convolution1DLayer(ConvolutionNDLayer):
    def __init__(self, kernel_shape, mode, input_shape=None):
        ConvolutionNDLayer.__init__(self, kernel_shape, mode, 1, input_shape)


class Convolution2DLayer(ConvolutionNDLayer):
    def __init__(self, kernel_shape, mode, input_shape=None):
        ConvolutionNDLayer.__init__(self, kernel_shape, mode, 2, input_shape)


class Convolution3DLayer(ConvolutionNDLayer):
    def __init__(self, kernel_shape, mode, input_shape=None):
        ConvolutionNDLayer.__init__(self, kernel_shape, mode, 3, input_shape)


convolution_map = {
    'nd': ConvolutionNDLayer,
    '1d': Convolution1DLayer,
    '2d': Convolution2DLayer,
    '3d': Convolution3DLayer,
}


def register_convolution(name: str, convolution: ConvolutionLayer):
    convolution_map[name.lower()] = convolution


class Convolution:
    def __init__(self, name: str, *args, **kwargs):
        self.__name = name.lower()
        self.__convolution = None
        if self.__name in convolution_map:
            self.__convolution = convolution_map[self.__name](*args, **kwargs)
        else:
            raise ValueError('No such convolution: {}'.format(name))

    def convolution_layer(self):
        return self.__convolution


class PoolingLayer(ConnectionLayer):
    def __init__(self, size: tuple, step: tuple, input_shape: tuple=None):
        ConnectionLayer.__init__(self, None, input_shape)
        self.size = size
        self.step = step

    def connection(self, input_symbol: V):
        pass

    def variables(self):
        return []

    def pooling_layer(self):
        return self


class MaxPoolingNDLayer(PoolingLayer):
    def __init__(self, size: tuple, step: tuple, dimension: int, input_shape: tuple=None):
        PoolingLayer.__init__(self, size, step, input_shape)
        self.__dimension = dimension
        self.__output_symbol = None

    def get_output_shape(self):
        return pooling_nd_shape(self.input_shape, self.size, self.step, self.__dimension)[0]

    output_shape = property(get_output_shape, ConnectionLayer.set_output_shape)

    def connection(self, input_symbol: V):
        if self.__output_symbol is None:
            self.__output_symbol = max_pooling_nd(input_symbol, self.size, self.step, self.__dimension)
        return self.__output_symbol


class MaxPooling1DLayer(MaxPoolingNDLayer):
    def __init__(self, size: tuple, step: tuple, input_shape: tuple=None):
        MaxPoolingNDLayer.__init__(self, size, step, 1, input_shape)


class MaxPooling2DLayer(MaxPoolingNDLayer):
    def __init__(self, size: tuple, step: tuple, input_shape: tuple=None):
        MaxPoolingNDLayer.__init__(self, size, step, 2, input_shape)


class MaxPooling3DLayer(MaxPoolingNDLayer):
    def __init__(self, size: tuple, step: tuple, input_shape: tuple=None):
        MaxPoolingNDLayer.__init__(self, size, step, 3, input_shape)


class AveragePoolingNDLayer(PoolingLayer):
    def __init__(self, size: tuple, step: tuple, dimension: int, input_shape: tuple=None):
        PoolingLayer.__init__(self, size, step, input_shape)
        self.__dimension = dimension
        self.__output_symbol = None

    def get_output_shape(self):
        return pooling_nd_shape(self.input_shape, self.size, self.step, self.__dimension)[0]

    output_shape = property(get_output_shape, ConnectionLayer.set_output_shape)

    def connection(self, input_symbol: V):
        if self.__output_symbol is None:
            self.__output_symbol = average_pooling_nd(input_symbol, self.size, self.step, self.__dimension)
        return self.__output_symbol


class AveragePooling1DLayer(AveragePoolingNDLayer):
    def __init__(self, size: tuple, step: tuple, input_shape: tuple = None):
        AveragePoolingNDLayer.__init__(self, size, step, 1, input_shape)


class AveragePooling2DLayer(AveragePoolingNDLayer):
    def __init__(self, size: tuple, step: tuple, input_shape: tuple = None):
        AveragePoolingNDLayer.__init__(self, size, step, 2, input_shape)


class AveragePooling3DLayer(AveragePoolingNDLayer):
    def __init__(self, size: tuple, step: tuple, input_shape: tuple = None):
        AveragePoolingNDLayer.__init__(self, size, step, 3, input_shape)


pooling_map = {
    'max_nd': MaxPoolingNDLayer,
    'max_1d': MaxPooling1DLayer,
    'max_2d': MaxPooling2DLayer,
    'max_3d': MaxPooling3DLayer,
    'average_nd': AveragePoolingNDLayer,
    'average_1d': AveragePooling1DLayer,
    'average_2d': AveragePooling2DLayer,
    'average_3d': AveragePooling3DLayer,
}


def register_pooling(name: str, pooling: PoolingLayer):
    pooling_map[name.lower()] = pooling


class Pooling:
    def __init__(self, name: str, *args, **kwargs):
        self.__name = name.lower()
        self.__pooling = None
        if self.__name in pooling_map:
            self.__pooling = pooling_map[self.__name](*args, **kwargs)
        else:
            raise ValueError('No such pooling: {}'.format(name))

    def pooling_layer(self):
        return self.__pooling


class UnpoolingLayer(ConnectionLayer):
    def __init__(self, size: tuple, step: tuple, input_shape: tuple=None):
        ConnectionLayer.__init__(self, None, input_shape)
        self.size = size
        self.step = step

    def unpooling_layer(self):
        return self

    def connection(self, input_symbol: V):
        pass

    def variables(self):
        return []


class AverageUnpoolingNDLayer(UnpoolingLayer):
    def __init__(self, size: tuple, step: tuple, dimension: int, input_shape: tuple=None):
        UnpoolingLayer.__init__(self, size, step, input_shape)
        self.__dimension = dimension
        self.__output_symbol = None

    def unpooling_function(self):
        return lambda data, pooling, size, step: average_unpooling_nd(pooling, size, step, self.__dimension)

    def get_output_shape(self):
        return unpooling_nd_shape(self.input_shape, self.size, self.step, None, self.__dimension)[0]

    def connection(self, input_symbol: V):
        if self.__output_symbol is None:
            self.__output_symbol = average_unpooling_nd(input_symbol, self.size, self.step, self.__dimension)
        return self.__output_symbol


class AverageUnpooling1DLayer(AverageUnpoolingNDLayer):
    def __init__(self, size: tuple, step: tuple, input_shape: tuple=None):
        AverageUnpoolingNDLayer.__init__(self, size, step, 1, input_shape)


class AverageUnpooling2DLayer(AverageUnpoolingNDLayer):
    def __init__(self, size: tuple, step: tuple, input_shape: tuple=None):
        AverageUnpoolingNDLayer.__init__(self, size, step, 2, input_shape)


class AverageUnpooling3DLayer(AverageUnpoolingNDLayer):
    def __init__(self, size: tuple, step: tuple, input_shape: tuple=None):
        AverageUnpoolingNDLayer.__init__(self, size, step, 3, input_shape)


unpooling_map = {
    'average_nd': AverageUnpoolingNDLayer,
    'average_1d': AverageUnpooling1DLayer,
    'average_2d': AverageUnpooling2DLayer,
    'average_3d': AverageUnpooling3DLayer,
}


def register_unpooling(name: str, unpooling: UnpoolingLayer):
    unpooling_map[name.lower()] = unpooling


class Unpooling:
    def __init__(self, name: str, *args, **kwargs):
        self.__name = name.lower()
        self.__unpooling = None
        if self.__name in unpooling_map:
            self.__unpooling = unpooling_map[self.__name](*args, **kwargs)
        else:
            raise ValueError('No such unpooling: {}'.format(name))

    def unpooling_layer(self):
        return self.__unpooling



import numpy as np
import matplotlib.pyplot as plt
import trik
# 每类随机生成点的个数。
points_sum = 100
c1_x = []
c1_y = []
c2_x = []
c2_y = []

# 分别在(0, 0)点附近和(8, 8)点附近生成2类随机数据。
for _ in range(points_sum):
    c1_x.append(np.random.normal(0, 2))
    c1_y.append(np.random.normal(0, 2))
    c2_x.append(np.random.normal(8, 2))
    c2_y.append(np.random.normal(8, 2))
c1 = trik.constant([c1_x, c1_y], name='c1')
c2 = trik.constant([c2_x, c2_y], name='c2')
W = trik.V([[1, 1], [1, 1]], name='w')
B = trik.V([[1], [1]], name='b')
# 定义SVM loss函数。
# loss = trik.reduce_mean(trik.maximum(0, [[1, -1]] @ (W @ c1 + B) + 1) + trik.maximum(0, [[-1, 1]] @ (W @ c2 + B) + 1))
loss1=trik.constant([[1,-1]])(trik.MM(),W(trik.MM(),c1)(trik.Plus(),B))(trik.Plus(),1)(trik.Max(),0)
loss2=trik.constant([[-1,1]])(trik.MM(),W(trik.MM(),c2)(trik.Plus(),B))(trik.Plus(),1)(trik.Max(),0)
loss=loss1(trik.Plus(),loss2)(trik.ReduceMean())
loss.set_variables([W,B])
optimizer = trik.SGD()
for epoch in range(1000):
    optimizer.minimize(loss)
    loss_value = loss.val()
    print('loss = {:.8f}'.format(loss_value))
    if loss_value < 0.00001:  # loss阈值。
        break
# 获取W和B的训练结果。
w_data = W.value
b_data = B.value
print(w_data,b_data)
# 计算分类直线的斜率和截距。
k = (w_data[1, 0] - w_data[0, 0]) / (w_data[0, 1] - w_data[1, 1])
b = (b_data[1, 0] - b_data[0, 0]) / (w_data[0, 1] - w_data[1, 1])
print(k,b)
# 分类面的端点。
x_range = np.array([np.min(c1_x), np.max(c2_x)])

plt.title('Paradox implement Linear SVM')
plt.plot(c1_x, c1_y, 'ro', label='Category 1')
plt.plot(c2_x, c2_y, 'bo', label='Category 2')
plt.plot(x_range, k * x_range + b, 'y', label='SVM')
plt.legend()
plt.show()



# -*- coding:utf-8 -*-
import trik

# 定义符号，A为方程系数矩阵，x为自变量，b为常数项。
A = trik.constant([[1, 2], [1, 3]], name='A')
x = trik.V([1, 0], name='x')
B = trik.constant([3, 4], name='b')

# 使用最小二乘误差定义loss。
# loss = trik.reduce_mean((A @ x - B) ** 2)
loss=A(trik.MM(),x)(trik.Sub(),B)(trik.Pow(),2)(trik.ReduceMean())#trik创建loss的计算引擎，申明变量为x。
loss.set_variables([x])

optimizer = trik.SGD()
for epoch in range(1000):
	optimizer.minimize(loss)
	loss_value = loss.val()
	if loss_value < 0.0000001:  # loss阈值。
		break
print('\nx={}'.format(x.value))
#~ print(loss.val())


from .tensor import V,as_symbols,element_wise_shape,numpy,reduce,Flip,Rotate90,ReduceMean,Expand,Operator
from ..yu import array_index_traversal, multi_range
from enum import Enum
#compute-------------
class ConvolutionMode(Enum):
	valid = 0
	full = 1
	
convolution_map = {
	'valid': ConvolutionMode.valid,
	'full': ConvolutionMode.full,
}

def __get_convolution_mode_string(mode):
	if isinstance(mode, str):
		if mode in convolution_map:
			return mode
		else:
			raise ValueError('No such convolution mode: {}'.format(mode))
	elif isinstance(mode, ConvolutionMode):
		return mode.name
	else:
		raise ValueError('Invalid mode type: {}'.format(type(mode)))


def basic_convolution_shape(shape_data, shape_kernel, dimension: int, mode: str):
	if mode == 'valid':
		return tuple(shape_data[i] - shape_kernel[i] + 1 for i in range(-dimension, 0))
	elif mode == 'full':
		return tuple(shape_data[i] + shape_kernel[i] - 1 for i in range(-dimension, 0))
	else:
		raise ValueError('Invalid convolution mode: {}'.format(mode))

def __compute_valid_convolution_nd(data, kernel, dimension: int):
	convolution_shape = tuple(data.shape[i] - kernel.shape[i] + 1 for i in range(-1, -dimension - 1, -1))
	list_dimension = reduce(lambda a, b: a * b, convolution_shape)
	data_prefix = data.shape[:-dimension]
	kernel_flat = kernel.ravel()
	data_flat = numpy.zeros(data_prefix + (list_dimension, len(kernel_flat)))
	for i in range(list_dimension):
		tensor_slice_start = [0] * len(kernel.shape)
		tensor_slice = [slice(None)] * len(data.shape)
		tensor_slice_start[-1] = i
		for r in range(-1, -len(kernel.shape) - 1, -1):
			dimension_scale = data.shape[r] - kernel.shape[r] + 1
			if tensor_slice_start[r] >= dimension_scale:
				tensor_slice_start[r + 1] = tensor_slice_start[r] // dimension_scale
				tensor_slice_start[r] %= dimension_scale
			tensor_slice[r] = slice(tensor_slice_start[r], tensor_slice_start[r] + kernel.shape[r])
		sub_convolution_index = (slice(None),) * (len(data.shape) - dimension) + tuple([i, slice(None)])
		data_flat[tuple(sub_convolution_index)] = data[tuple(tensor_slice)].reshape(data_prefix + (reduce(lambda a, b: a * b, kernel.shape),))
	convolution_flat = numpy.matmul(data_flat, numpy.flip(kernel_flat, axis=0))
	convolution_nd = convolution_flat.reshape(data_prefix + convolution_shape)
	return convolution_nd

def __compute_convolution_nd(data, kernel, dimension: int, mode: str):
	mode_string = __get_convolution_mode_string(mode)
	for i in range(dimension):
		if data.shape[i] < kernel.shape[i]:
			raise ValueError('Data shape smaller than kernel shape: {} {}'.format(data.shape, kernel.shape))
	if mode_string == 'valid':
		return __compute_valid_convolution_nd(data, kernel, dimension)
	elif mode_string == 'full':
		data_prefix = data.shape[:-dimension]
		expand_data = numpy.zeros(data_prefix + tuple(data.shape[i] + (kernel.shape[i] - 1) * 2 for i in range(dimension)))
		padding = tuple(kernel.shape[i] - 1 for i in range(dimension))
		expand_data[tuple(slice(None) for _ in data_prefix) + tuple(slice(padding[i], padding[i] + data.shape[i]) for i in range(dimension))] = data
		return __compute_valid_convolution_nd(expand_data, kernel, dimension)
	else:
		raise ValueError('Never reached.')


def compute_convolution_nd(data, kernel, dimension: int, mode=ConvolutionMode.valid, element_wise: bool=False):
	mode_string = __get_convolution_mode_string(mode)
	result = []
	data_prefix_shape = data.shape[:-dimension]
	kernel_prefix_shape = kernel.shape[:-dimension]
	if element_wise:
		final_shape = element_wise_shape(data_prefix_shape, kernel_prefix_shape)[0]
		data = numpy.broadcast_to(data, final_shape + data.shape[-2:])
		kernel = numpy.broadcast_to(kernel, final_shape + kernel.shape[-2:])
		if final_shape:
			for index in array_index_traversal(final_shape):
				result.append(__compute_convolution_nd(data[index], kernel[index], dimension, mode_string))
			return numpy.array(result).reshape(final_shape + result[0].shape)
		else:
			return __compute_convolution_nd(data, kernel, dimension, mode_string)
	else:
		if kernel_prefix_shape:
			final_shape = data_prefix_shape + kernel_prefix_shape + basic_convolution_shape(data.shape[-dimension:], kernel.shape[-dimension:], dimension, mode_string)
			result = numpy.zeros(final_shape)
			for kernel_index in array_index_traversal(kernel_prefix_shape):
				sub_result_index = tuple(slice(None) for _ in data_prefix_shape) + kernel_index + tuple(slice(None) for _ in range(dimension))
				result[sub_result_index] = __compute_convolution_nd(data, kernel[kernel_index], dimension, mode_string)
			return result
		else:
			return __compute_convolution_nd(data, kernel, dimension, mode_string)

def __compute_max_pooling_nd(data, size, step, dimension: int, reference=None):
	for i in range(dimension):
		if data.shape[i] < size[i]:
			raise ValueError('Data shape smaller than size: {} {}'.format(data.shape, size))
	pooling_array = []
	pooling_grid = [range(0, data.shape[i] - size[i] + 1, step[i]) for i in range(dimension)]
	for index in multi_range(pooling_grid):
		sub_slice = [slice(index[i], index[i] + size[i]) for i in range(dimension)]
		if reference is None:
			pooling_array.append(numpy.max(data[tuple(sub_slice)]))
		else:
			max_index = numpy.argmax(reference[sub_slice])
			sub_data = data[sub_slice]
			pooling_array.append(sub_data[numpy.unravel_index(max_index, sub_data.shape)])
	
	return numpy.array(pooling_array).reshape([len(g) for g in pooling_grid])

def compute_max_pooling_nd(data, size, step, dimension: int, reference=None):
	result = []
	data_prefix_shape = data.shape[:-dimension]
	if data_prefix_shape:
		for key in array_index_traversal(data_prefix_shape):
			if reference is None:
				result.append(__compute_max_pooling_nd(data[key], size, step, dimension))
			else:
				result.append(__compute_max_pooling_nd(data[key], size, step, dimension, reference[key]))
		return numpy.array(result).reshape(data_prefix_shape + result[0].shape)
	else:
		return __compute_max_pooling_nd(data, size, step, dimension)


def __compute_max_unpooling_nd(data, pooling, size, step, dimension: int):
	for i in range(dimension):
		if data.shape[i] < size[i]:
			raise ValueError('Data shape smaller than size: {} {}'.format(data.shape, size))
	unpooling_array = numpy.zeros(data.shape)
	unpooling_grid = [range(0, data.shape[i] - size[i] + 1, step[i]) for i in range(dimension)]
	for n, index in enumerate(multi_range(unpooling_grid)):
		sub_slice = tuple([slice(index[i], index[i] + size[i]) for i in range(dimension)])
		max_index = numpy.argmax(data[sub_slice])
		sub_unpooling_array = unpooling_array[sub_slice]
		sub_unpooling_array[numpy.unravel_index(max_index, sub_unpooling_array.shape)] = pooling[numpy.unravel_index(n, pooling.shape)]
	return unpooling_array


def compute_max_unpooling_nd(data, pooling, size, step, dimension: int):
	result = []
	data_prefix_shape = data.shape[:-dimension]
	kernel_prefix_shape = pooling.shape[:-dimension]
	final_shape = element_wise_shape(data_prefix_shape, kernel_prefix_shape)[0]
	data = numpy.broadcast_to(data, final_shape + data.shape[-dimension:])
	pooling = numpy.broadcast_to(pooling, final_shape + pooling.shape[-dimension:])
	if final_shape:
		for key in array_index_traversal(final_shape):
			result.append(__compute_max_unpooling_nd(data[key], pooling[key], size, step, dimension))
		return numpy.array(result).reshape(final_shape + result[0].shape)
	else:
		return __compute_max_unpooling_nd(data, pooling, size, step, dimension)


def __compute_average_pooling_nd(data, size, step, dimension: int):
	for i in range(dimension):
		if data.shape[i] < size[i]:
			raise ValueError('Data shape smaller than size: {} {}'.format(data.shape, size))
	pooling_array = []
	pooling_grid = [range(0, data.shape[i] - size[i] + 1, step[i]) for i in range(dimension)]
	for index in multi_range(pooling_grid):
		pooling_array.append(numpy.mean(data[tuple([slice(index[i], index[i] + size[i]) for i in range(dimension)])]))
	return numpy.array(pooling_array).reshape([len(g) for g in pooling_grid])


def compute_average_pooling_nd(data, size, step, dimension: int):
	result = []
	data_prefix_shape = data.shape[:-dimension]
	if data_prefix_shape:
		for key in array_index_traversal(data_prefix_shape):
			result.append(__compute_average_pooling_nd(data[key], size, step, dimension))
		return numpy.array(result).reshape(data_prefix_shape + result[0].shape)
	else:
		return __compute_average_pooling_nd(data, size, step, dimension)


def __compute_average_unpooling_nd(pooling, size, step, dimension: int, unpooling_size=None):
	if unpooling_size is None:
		unpooling_array = numpy.zeros([size[i] + (pooling.shape[i] - 1) * step[i] for i in range(dimension)])
	else:
		unpooling_array = numpy.zeros(unpooling_size)
	unpooling_grid = [range(0, unpooling_array.shape[i] - size[i] + 1, step[i]) for i in range(dimension)]
	for n, index in enumerate(multi_range(unpooling_grid)):
		sub_slice = tuple(slice(index[i], index[i] + size[i]) for i in range(dimension))
		unpooling_array[sub_slice] += pooling[numpy.unravel_index(n, pooling.shape)]
	return unpooling_array


def compute_average_unpooling_nd(pooling, size, step, dimension: int, unpooling_size=None):
	result = []
	data_prefix_shape = pooling.shape[:-dimension]
	if data_prefix_shape:
		for key in array_index_traversal(data_prefix_shape):
			result.append(__compute_average_unpooling_nd(pooling[key], size, step, dimension, unpooling_size))
		return numpy.array(result).reshape(data_prefix_shape + result[0].shape)
	else:
		return __compute_average_unpooling_nd(pooling, size, step, dimension, unpooling_size)

#function----------

def convolution_nd(data, kernel, dimension: int, mode, element_wise: bool=False):
	return V(operator=ConvolutionND(dimension, mode, element_wise), inputs=as_symbols([data, kernel]))


def convolution_1d(data, kernel, mode, element_wise: bool=False):
	return convolution_nd(data, kernel, 1, mode, element_wise)


def convolution_2d(data, kernel, mode, element_wise: bool=False):
	return convolution_nd(data, kernel, 2, mode, element_wise)


def convolution_3d(data, kernel, mode, element_wise: bool=False):
	return convolution_nd(data, kernel, 3, mode, element_wise)


def max_pooling_nd(data, size: tuple, step: tuple, dimension: int, reference=None):
	if reference is None:
		return V(operator=MaxPoolingND(dimension, size, step), inputs=as_symbols([data]))
	else:
		return V(operator=MaxReferencePoolingND(dimension, size, step), inputs=as_symbols([data, reference]))


def max_pooling_1d(data, size: tuple, step: tuple, reference=None):
	return max_pooling_nd(data, size, step, 1, reference)


def max_pooling_2d(data, size: tuple, step: tuple, reference=None):
	return max_pooling_nd(data, size, step, 2, reference)


def max_pooling_3d(data, size: tuple, step: tuple, reference=None):
	return max_pooling_nd(data, size, step, 3, reference)


def max_unpooling_nd(data, pooling, size: tuple, step: tuple, dimension: int):
	return V(operator=MaxUnpoolingND(dimension, size, step), inputs=as_symbols([data, pooling]))


def max_unpooling_1d(data, pooling, size: tuple, step: tuple):
	return max_unpooling_nd(data, pooling, size, step, 1)


def max_unpooling_2d(data, pooling, size: tuple, step: tuple):
	return max_unpooling_nd(data, pooling, size, step, 2)


def max_unpooling_3d(data, pooling, size: tuple, step: tuple):
	return max_unpooling_nd(data, pooling, size, step, 3)


def average_pooling_nd(data, size: tuple, step: tuple, dimension: int):
	return V(operator=AveragePoolingND(dimension, size, step), inputs=as_symbols([data]))


def average_pooling_1d(data, size: tuple, step: tuple):
	return average_pooling_nd(data, size, step, 1)


def average_pooling_2d(data, size: tuple, step: tuple):
	return average_pooling_nd(data, size, step, 2)


def average_pooling_3d(data, size: tuple, step: tuple):
	return average_pooling_nd(data, size, step, 3)


def average_unpooling_nd(pooling, size: tuple, step: tuple, dimension: int, unpooling_size: int=None):
	return V(operator=AverageUnpoolingND(dimension, size, step, unpooling_size), inputs=as_symbols([pooling]))


def average_unpooling_1d(pooling, size: tuple, step: tuple, unpooling_size: tuple=None):
	return average_unpooling_nd(pooling, size, step, 1, unpooling_size)


def average_unpooling_2d(pooling, size: tuple, step: tuple, unpooling_size: tuple=None):
	return average_unpooling_nd(pooling, size, step, 2, unpooling_size)


def average_unpooling_3d(pooling, size: tuple, step: tuple, unpooling_size: tuple=None):
	return average_unpooling_nd(pooling, size, step, 3, unpooling_size)

#operator---------------
def convolution_nd_shape(shape_data, shape_kernel, dimension, mode):
	prefix_shape = shape_data[:-dimension] + shape_kernel[:-dimension]
	if mode == 'valid' or mode == ConvolutionMode.valid:
		new_shape = prefix_shape + tuple(shape_data[i] - shape_kernel[i] + 1 for i in range(-dimension, 0))
	elif mode == 'full' or mode == ConvolutionMode.full:
		new_shape = prefix_shape + tuple(shape_data[i] + shape_kernel[i] - 1 for i in range(-dimension, 0))
	else:
		raise ValueError('Invalid convolution mode: {}'.format(mode))
	return new_shape, (), ()


def element_wise_convolution_nd_shape(shape_data, shape_kernel, dimension, mode):
	prefix_shape, prefix_broadcast_data, prefix_broadcast_kernel = element_wise_shape(shape_data[:-dimension], shape_kernel[:-dimension])
	if mode == 'valid' or mode == ConvolutionMode.valid:
		new_shape = prefix_shape + tuple(shape_data[i] - shape_kernel[i] + 1 for i in range(-dimension, 0))
	elif mode == 'full' or mode == ConvolutionMode.full:
		new_shape = prefix_shape + tuple(shape_data[i] + shape_kernel[i] - 1 for i in range(-dimension, 0))
	else:
		raise ValueError('Invalid convolution mode: {}'.format(mode))
	return new_shape, prefix_broadcast_data + (0,) * dimension, prefix_broadcast_kernel + (0,) * dimension


def pooling_nd_shape(shape_data, size, step, dimension):
	prefix_shape = shape_data[:-dimension]
	if not isinstance(size, tuple):
		size = (size,)
	if not isinstance(step, tuple):
		step = (step,)
	new_shape = prefix_shape + tuple(len(range(0, shape_data[i] - size[i] + 1, step[i])) for i in range(-dimension, 0))
	return new_shape, ()


def max_unpooling_nd_shape(shape_data, shape_pooling, dimension):
	prefix_shape, prefix_broadcast_data, prefix_broadcast_kernel = element_wise_shape(shape_data[:-dimension], shape_pooling[:-dimension])
	new_shape = prefix_shape + tuple(shape_data[i] for i in range(-dimension, 0))
	return new_shape, prefix_broadcast_data + (0,) * dimension, prefix_broadcast_kernel + (0,) * dimension


def unpooling_nd_shape(shape_pooling, size, step, unpooling_size, dimension):
	prefix_shape, prefix_broadcast_data, prefix_broadcast_kernel = element_wise_shape(shape_pooling[:-dimension], shape_pooling[:-dimension])
	new_shape = prefix_shape + tuple((size[i] + (shape_pooling[i] - 1) * step[i]) if unpooling_size is None else unpooling_size[i] for i in range(-dimension, 0))
	return new_shape, prefix_broadcast_data + (0,) * dimension, prefix_broadcast_kernel + (0,) * dimension


class ConvolutionND(Operator):
	def __init__(self, dimension: int, mode, element_wise: bool=False):
		self.inputs_count = 2
		self.arguments = {'dimension': dimension, 'mode': mode, 'element_wise': element_wise}
		assert dimension > 0

	def f(self, value_data, value_kernel):
		res=compute_convolution_nd(value_data, value_kernel, **self.arguments)
		return res

	def bprop(self, engine, symbol_forward, symbol_data, symbol_kernel):
		forward = engine.bprop(symbol_forward)
		dimension = self.arguments['dimension']
		mode = self.arguments['mode']
		if mode == 'valid' or mode == ConvolutionMode.valid:
			prefix_shape_kernel = engine.shape2(symbol_kernel)[:-dimension]
			prefix_shape_data = engine.shape2(symbol_data)[:-dimension]
			if dimension == 2:
				flip_kernel = symbol_kernel(Rotate90(count=2, axes=(-2, -1)))
			else:
				flip_kernel = symbol_kernel
				for i in range(dimension):
					flip_kernel = flip_kernel(Flip(-1 - i))
			gradient_data = convolution_nd(forward, flip_kernel, dimension, ConvolutionMode.full, True)
			for _ in prefix_shape_kernel:
				gradient_data = gradient_data(ReduceMean(axis=-dimension - 1))
				symbol_data = symbol_data(Expand(-dimension - 1))
			gradient_kernel = convolution_nd(symbol_data, forward, dimension, ConvolutionMode.valid, True)
			for _ in prefix_shape_data:
				gradient_kernel = gradient_kernel(ReduceMean(axis=-dimension - 1 - len(prefix_shape_kernel)))
			return [lambda: gradient_data,
					lambda: gradient_kernel]
		elif mode == 'full' or mode == ConvolutionMode.full:
			prefix_shape_kernel = engine.shape2(symbol_kernel)[:-dimension]
			prefix_shape_data = engine.shape2(symbol_data)[:-dimension]
			if dimension == 2:
				flip_kernel = symbol_kernel(Rotate90(count=2, axes=(-2, -1)))
			else:
				flip_kernel = symbol_kernel
				for i in range(dimension):
					flip_kernel = flip_kernel(Flip(-1 - i))
			gradient_data = convolution_nd(forward, flip_kernel, dimension, ConvolutionMode.valid, True)
			for _ in prefix_shape_kernel:
				gradient_data = gradient_data(ReduceMean(axis=-dimension - 1))
				symbol_data = symbol_data(Expand(-dimension - 1))
			flip_gradient = convolution_nd(forward, symbol_data, dimension, ConvolutionMode.valid, True)
			if dimension == 2:
				gradient_kernel = flip_gradient(Rotate90(count=2, axes=(-2, -1)))
			else:
				gradient_kernel = flip_gradient
				for i in range(dimension):
					gradient_kernel = gradient_kernel(Flip(-1 - i))
			for _ in prefix_shape_data:
				gradient_kernel = gradient_kernel(ReduceMean(axis=-dimension - 1 - len(prefix_shape_kernel)))
			return [lambda: gradient_data,
					lambda: gradient_kernel]
		else:
			raise ValueError('Invalid mode: {}'.format(mode))

	def shape(self, shape_data, shape_kernel):
		if self.arguments['element_wise']:
			return element_wise_convolution_nd_shape(shape_data, shape_kernel, self.arguments['dimension'], self.arguments['mode'])
		else:
			return convolution_nd_shape(shape_data, shape_kernel, self.arguments['dimension'], self.arguments['mode'])


class Convolution1D(ConvolutionND):
	def __init__(self, mode, element_wise: bool=False):
		ConvolutionND.__init__(self, 1, mode, element_wise)


class Convolution2D(ConvolutionND):
	def __init__(self, mode, element_wise: bool=False):
		ConvolutionND.__init__(self, 2, mode, element_wise)


class Convolution3D(ConvolutionND):
	def __init__(self, mode, element_wise: bool=False):
		ConvolutionND.__init__(self, 3, mode, element_wise)


class MaxPoolingND(Operator):
	def __init__(self, dimension: int, size: tuple, step: tuple):
		self.inputs_count = 1
		self.arguments = {'dimension': dimension, 'size': size, 'step': step}

	def f(self, value_data):
		res=compute_max_pooling_nd(value_data, **self.arguments)
		return res

	def bprop(self, engine, symbol_forward, symbol_data):
		forward = engine.bprop(symbol_forward)
		return [lambda: max_unpooling_nd(symbol_data, forward, **self.arguments)]

	def shape(self, shape_data):
		return pooling_nd_shape(shape_data, **self.arguments)


class MaxPooling1D(MaxPoolingND):
	def __init__(self, size: tuple, step: tuple):
		MaxPoolingND.__init__(self, 1, size, step)


class MaxPooling2D(MaxPoolingND):
	def __init__(self, size: tuple, step: tuple):
		MaxPoolingND.__init__(self, 2, size, step)


class MaxPooling3D(MaxPoolingND):
	def __init__(self, size: tuple, step: tuple):
		MaxPoolingND.__init__(self, 3, size, step)


class MaxReferencePoolingND(Operator):
	def __init__(self, dimension: int, size: tuple, step: tuple):
		self.inputs_count = 2
		self.arguments = {'dimension': dimension, 'size': size, 'step': step}

	def f(self, value_data, reference_data):
		res=compute_max_pooling_nd(value_data, reference=reference_data, **self.arguments)
		return res

	def bprop(self, engine, symbol_forward, symbol_data, symbol_reference):
		forward = engine.bprop(symbol_forward)
		return [lambda: max_unpooling_nd(symbol_reference, forward, **self.arguments),
				lambda: max_unpooling_nd(symbol_reference, numpy.ones(engine.shape2(forward)), **self.arguments)]

	def shape(self, shape_data):
		return pooling_nd_shape(shape_data, **self.arguments)


class MaxReferencePooling1D(MaxReferencePoolingND):
	def __init__(self, size: tuple, step: tuple):
		MaxReferencePoolingND.__init__(self, 1, size, step)


class MaxReferencePooling2D(MaxReferencePoolingND):
	def __init__(self, size: tuple, step: tuple):
		MaxReferencePoolingND.__init__(self, 2, size, step)


class MaxReferencePooling3D(MaxReferencePoolingND):
	def __init__(self, size: tuple, step: tuple):
		MaxReferencePoolingND.__init__(self, 3, size, step)


class MaxUnpoolingND(Operator):
	def __init__(self, dimension: int, size: tuple, step: tuple):
		self.inputs_count = 2
		self.arguments = {'dimension': dimension, 'size': size, 'step': step}

	def f(self, value_data, value_pooling):
		res=compute_max_unpooling_nd(value_data, value_pooling, **self.arguments)
		return res

	def bprop(self, engine, symbol_forward, symbol_data, symbol_pooling):
		forward = engine.bprop(symbol_forward)
		return [lambda: max_unpooling_nd(symbol_data, numpy.ones(engine.shape2(symbol_pooling)), **self.arguments),
				lambda: max_pooling_nd(forward, reference=symbol_data, **self.arguments)]

	def shape(self, shape_data, shape_pooling):
		return max_unpooling_nd_shape(shape_data, shape_pooling, self.arguments['dimension'])


class MaxUnpooling1D(MaxUnpoolingND):
	def __init__(self, size: tuple, step: tuple):
		MaxUnpoolingND.__init__(self, 1, size, step)


class MaxUnpooling2D(MaxUnpoolingND):
	def __init__(self, size: tuple, step: tuple):
		MaxUnpoolingND.__init__(self, 2, size, step)


class MaxUnpooling3D(MaxUnpoolingND):
	def __init__(self, size: tuple, step: tuple):
		MaxUnpoolingND.__init__(self, 3, size, step)


class AveragePoolingND(Operator):
	def __init__(self, dimension: int, size: tuple, step: tuple):
		self.inputs_count = 1
		self.arguments = {'dimension': dimension, 'size': size, 'step': step}

	def f(self, value_data):
		return compute_average_pooling_nd(value_data, **self.arguments)

	def bprop(self, engine, symbol_forward, symbol_data):
		forward = engine.bprop(symbol_forward)
		return [lambda: average_unpooling_nd(forward, unpooling_size=engine.shape2(symbol_data)[-self.arguments['dimension']:], **self.arguments)]

	def shape(self, shape_data):
		return pooling_nd_shape(shape_data, **self.arguments)


class AveragePooling1D(AveragePoolingND):
	def __init__(self, size: tuple, step: tuple):
		AveragePoolingND.__init__(self, 1, size, step)


class AveragePooling2D(AveragePoolingND):
	def __init__(self, size: tuple, step: tuple):
		AveragePoolingND.__init__(self, 2, size, step)


class AveragePooling3D(AveragePoolingND):
	def __init__(self, size: tuple, step: tuple):
		AveragePoolingND.__init__(self, 3, size, step)


class AverageUnpoolingND(Operator):
	def __init__(self, dimension: int, size: tuple, step: tuple, unpooling_size: tuple=None):
		self.inputs_count = 1
		self.arguments = {'dimension': dimension, 'size': size, 'step': step, 'unpooling_size': unpooling_size}

	def f(self, value_pooling):
		return compute_average_unpooling_nd(value_pooling, **self.arguments)

	def bprop(self, engine, symbol_forward, symbol_pooling):
		forward = engine.bprop(symbol_forward)
		return [lambda: average_pooling_nd(forward, **self.arguments)]

	def shape(self, shape_data):
		return unpooling_nd_shape(shape_data, **self.arguments)


class AverageUnpooling1D(AverageUnpoolingND):
	def __init__(self, size: tuple, step: tuple, unpooling_size: tuple=None):
		AverageUnpoolingND.__init__(self, 1, size, step, unpooling_size)


class AverageUnpooling2D(AverageUnpoolingND):
	def __init__(self, size: tuple, step: tuple, unpooling_size: tuple=None):
		AverageUnpoolingND.__init__(self, 2, size, step, unpooling_size)


class AverageUnpooling3D(AverageUnpoolingND):
	def __init__(self, size: tuple, step: tuple, unpooling_size: tuple=None):
		AverageUnpoolingND.__init__(self, 3, size, step, unpooling_size)
